"""Binary runner and results class."""
import os
import subprocess
import time
from pathlib import Path
from typing import List, Optional, Union


class SubprocessRunResults:
    """Results returned from subprocess.run()."""

    def __init__(
            self, stdout, stderr, return_code: int,
            process_time: Optional[float] = None):
        self.stdout = stdout
        self.stderr = stderr
        self.return_code = return_code
        self.success = return_code == 0
        self.process_time = process_time


class SimpleBinaryRunner:
    """Class to execute a subprocess."""
    path_type = Union[str, Path]

    def __init__(self,
                 binary,
                 run_argv: List[str],
                 omp_num_threads: int,
                 directory: path_type = './',
                 args=None) -> None:
        """Initialise class.

        :param binary: Binary name prepended by full path, or just binary name
            (if present in $PATH).
        :param run_argv: Run commands sequentially as a list of str.
            For example:
            * For serial: ['./'] or ['']
            * For MPI:   ['mpirun', '-np', '2']
        :param omp_num_threads: Number of OMP threads.
        :param args: Optional arguments for the binary.
        """
        if args is None:
            args = []
        self.binary = binary
        self.directory = directory

        self.run_argv = run_argv

        self.omp_num_threads = omp_num_threads
        self.args = args

        if directory is not None and not Path(directory).is_dir():
            raise OSError(f"Run directory does not exist: {directory}")

        if omp_num_threads <= 0:
            raise ValueError("Number of OMP threads must be > 0")

    def compose_execution_list(self) -> list:
        """Generate a complete list of strings to pass to subprocess.run().

        This is done to execute the calculation.

        For example, given:
          ['mpirun', '-np, '2'] + ['binary.exe'] + ['>', 'std.out']

        return ['mpirun', '-np, '2', 'binary.exe', '>', 'std.out']
        """
        return self.run_argv + [self.binary] + self.args

    def run(self) -> SubprocessRunResults:
        """Run a binary."""
        execution_list = self.compose_execution_list()
        my_env = {**os.environ}

        time_start: float = time.time()
        result = subprocess.run(execution_list,
                                env=my_env,
                                capture_output=True,
                                cwd=self.directory, check=False)
        total_time = time.time() - time_start
        return SubprocessRunResults(
            result.stdout, result.stderr, result.returncode, total_time)



import numpy as np
import pytest

from ase import Atoms
from ase.formula import Formula


def test_formula_things():
    assert Formula('A3B2C2D').format('abc') == 'DB2C2A3'
    assert Formula('A3B2C2D').format('ab2') == 'DB2C2A3'
    assert Formula('A3B2C2D').format('a2b') == 'A3B2C2D'
    assert str(Formula('HHOOO', format='reduce')) == 'H2O3'
    assert Formula('HHOOOUO').format('reduce') == 'H2O3UO'


@pytest.mark.parametrize('f',
                         ['SiC', 'MoS2', 'GaAs', 'CO', 'NH3'])
def test_periodic(f):
    for fmt in ['ab2', 'a2b', 'periodic']:
        assert Formula(f, format=fmt).format('periodic') == f


def test_atoms_formula_things():
    assert Atoms('MoS2').get_chemical_formula() == 'MoS2'
    assert Atoms('SnO2').get_chemical_formula(mode='metal') == 'SnO2'


def test_h0c1():
    f = Formula.from_dict({'H': 0, 'C': 1})
    assert f.format('hill') == 'C'
    with pytest.raises(ValueError):
        Formula.from_dict({'H': -1})
    with pytest.raises(ValueError):
        Formula.from_dict({'H': 1.5})
    with pytest.raises(ValueError):
        Formula.from_dict({7: 1})


def test_formula():
    for sym in ['', 'Pu', 'Pu2', 'U2Pu2', 'U2((Pu2)2H)']:
        for mode in ['all', 'reduce', 'hill', 'metal']:
            for empirical in [False, True]:
                if empirical and mode in ['all', 'reduce']:
                    continue
                atoms = Atoms(sym)
                formula = atoms.get_chemical_formula(mode=mode,
                                                     empirical=empirical)
                atoms2 = Atoms(formula)
                print(repr(sym), '->', repr(formula))
                n1 = np.sort(atoms.numbers)
                n2 = np.sort(atoms2.numbers)
                if empirical and len(atoms) > 0:
                    reduction = len(n1) // len(n2)
                    n2 = np.repeat(n2, reduction)
                assert (n1 == n2).all()


def test_convert():
    assert str(Formula('AgAg').convert('hill')) == 'Ag2'


def test_formula_on_formula():
    formula = Formula('CH3CH2OH')
    assert formula == Formula(formula)


@pytest.mark.parametrize(
    'x',
    ['H2O', '10H2O', '2(CuO2(H2O)2)10', 'Cu20+H2', 'H' * 15, 'AuBC2', ''])
def test_formulas(x):
    f = Formula(x)
    y = str(f)
    assert y == x
    print(f.count(), f'{f:latex}')
    a, b = divmod(f, 'H2O')
    assert a * Formula('H2O') + b == f
    assert f != 117  # check that formula can be compared to non-formula object



from typing import IO, Optional, Union

import numpy as np

from ase import Atoms
from ase.optimize.optimize import Optimizer
from ase.utils.linesearch import LineSearch


class LBFGS(Optimizer):
    """Limited memory BFGS optimizer.

    A limited memory version of the bfgs algorithm. Unlike the bfgs algorithm
    used in bfgs.py, the inverse of Hessian matrix is updated.  The inverse
    Hessian is represented only as a diagonal matrix to save memory

    """

    def __init__(
        self,
        atoms: Atoms,
        restart: Optional[str] = None,
        logfile: Union[IO, str] = '-',
        trajectory: Optional[str] = None,
        maxstep: Optional[float] = None,
        memory: int = 100,
        damping: float = 1.0,
        alpha: float = 70.0,
        use_line_search: bool = False,
        **kwargs,
    ):
        """

        Parameters
        ----------
        atoms: :class:`~ase.Atoms`
            The Atoms object to relax.

        restart: str
            JSON file used to store vectors for updating the inverse of
            Hessian matrix. If set, file with such a name will be searched
            and information stored will be used, if the file exists.

        logfile: file object or str
            If *logfile* is a string, a file with that name will be opened.
            Use '-' for stdout.

        trajectory: string
            Trajectory file used to store optimisation path.

        maxstep: float
            How far is a single atom allowed to move. This is useful for DFT
            calculations where wavefunctions can be reused if steps are small.
            Default is 0.2 Angstrom.

        memory: int
            Number of steps to be stored. Default value is 100. Three numpy
            arrays of this length containing floats are stored.

        damping: float
            The calculated step is multiplied with this number before added to
            the positions.

        alpha: float
            Initial guess for the Hessian (curvature of energy surface). A
            conservative value of 70.0 is the default, but number of needed
            steps to converge might be less if a lower value is used. However,
            a lower value also means risk of instability.

        kwargs : dict, optional
            Extra arguments passed to
            :class:`~ase.optimize.optimize.Optimizer`.

        """
        Optimizer.__init__(self, atoms, restart, logfile, trajectory, **kwargs)

        if maxstep is not None:
            self.maxstep = maxstep
        else:
            self.maxstep = self.defaults['maxstep']

        if self.maxstep > 1.0:
            raise ValueError('You are using a much too large value for ' +
                             'the maximum step size: %.1f Angstrom' %
                             self.maxstep)

        self.memory = memory
        # Initial approximation of inverse Hessian 1./70. is to emulate the
        # behaviour of BFGS. Note that this is never changed!
        self.H0 = 1. / alpha
        self.damping = damping
        self.use_line_search = use_line_search
        self.p = None
        self.function_calls = 0
        self.force_calls = 0

    def initialize(self):
        """Initialize everything so no checks have to be done in step"""
        self.iteration = 0
        self.s = []
        self.y = []
        # Store also rho, to avoid calculating the dot product again and
        # again.
        self.rho = []

        self.r0 = None
        self.f0 = None
        self.e0 = None
        self.task = 'START'
        self.load_restart = False

    def read(self):
        """Load saved arrays to reconstruct the Hessian"""
        self.iteration, self.s, self.y, self.rho, \
            self.r0, self.f0, self.e0, self.task = self.load()
        self.load_restart = True

    def step(self, forces=None):
        """Take a single step

        Use the given forces, update the history and calculate the next step --
        then take it"""

        if forces is None:
            forces = self.optimizable.get_forces()

        pos = self.optimizable.get_positions()

        self.update(pos, forces, self.r0, self.f0)

        s = self.s
        y = self.y
        rho = self.rho
        H0 = self.H0

        loopmax = np.min([self.memory, self.iteration])
        a = np.empty((loopmax,), dtype=np.float64)

        # ## The algorithm itself:
        q = -forces.reshape(-1)
        for i in range(loopmax - 1, -1, -1):
            a[i] = rho[i] * np.dot(s[i], q)
            q -= a[i] * y[i]
        z = H0 * q

        for i in range(loopmax):
            b = rho[i] * np.dot(y[i], z)
            z += s[i] * (a[i] - b)

        self.p = - z.reshape((-1, 3))
        # ##

        g = -forces
        if self.use_line_search is True:
            e = self.func(pos)
            self.line_search(pos, g, e)
            dr = (self.alpha_k * self.p).reshape(len(self.optimizable), -1)
        else:
            self.force_calls += 1
            self.function_calls += 1
            dr = self.determine_step(self.p) * self.damping
        self.optimizable.set_positions(pos + dr)

        self.iteration += 1
        self.r0 = pos
        self.f0 = -g
        self.dump((self.iteration, self.s, self.y,
                   self.rho, self.r0, self.f0, self.e0, self.task))

    def determine_step(self, dr):
        """Determine step to take according to maxstep

        Normalize all steps as the largest step. This way
        we still move along the eigendirection.
        """
        steplengths = (dr**2).sum(1)**0.5
        longest_step = np.max(steplengths)
        if longest_step >= self.maxstep:
            dr *= self.maxstep / longest_step

        return dr

    def update(self, pos, forces, r0, f0):
        """Update everything that is kept in memory

        This function is mostly here to allow for replay_trajectory.
        """
        if self.iteration > 0:
            s0 = pos.reshape(-1) - r0.reshape(-1)
            self.s.append(s0)

            # We use the gradient which is minus the force!
            y0 = f0.reshape(-1) - forces.reshape(-1)
            self.y.append(y0)

            rho0 = 1.0 / np.dot(y0, s0)
            self.rho.append(rho0)

        if self.iteration > self.memory:
            self.s.pop(0)
            self.y.pop(0)
            self.rho.pop(0)

    def replay_trajectory(self, traj):
        """Initialize history from old trajectory."""
        if isinstance(traj, str):
            from ase.io.trajectory import Trajectory
            traj = Trajectory(traj, 'r')
        r0 = None
        f0 = None
        # The last element is not added, as we get that for free when taking
        # the first qn-step after the replay
        for i in range(len(traj) - 1):
            pos = traj[i].get_positions()
            forces = traj[i].get_forces()
            self.update(pos, forces, r0, f0)
            r0 = pos.copy()
            f0 = forces.copy()
            self.iteration += 1
        self.r0 = r0
        self.f0 = f0

    def func(self, x):
        """Objective function for use of the optimizers"""
        self.optimizable.set_positions(x.reshape(-1, 3))
        self.function_calls += 1
        return self.optimizable.get_potential_energy()

    def fprime(self, x):
        """Gradient of the objective function for use of the optimizers"""
        self.optimizable.set_positions(x.reshape(-1, 3))
        self.force_calls += 1
        # Remember that forces are minus the gradient!
        return - self.optimizable.get_forces().reshape(-1)

    def line_search(self, r, g, e):
        self.p = self.p.ravel()
        p_size = np.sqrt((self.p**2).sum())
        if p_size <= np.sqrt(len(self.optimizable) * 1e-10):
            self.p /= (p_size / np.sqrt(len(self.optimizable) * 1e-10))
        g = g.ravel()
        r = r.ravel()
        ls = LineSearch()
        self.alpha_k, e, self.e0, self.no_update = \
            ls._line_search(self.func, self.fprime, r, self.p, g, e, self.e0,
                            maxstep=self.maxstep, c1=.23,
                            c2=.46, stpmax=50.)
        if self.alpha_k is None:
            raise RuntimeError('LineSearch failed!')


class LBFGSLineSearch(LBFGS):
    """This optimizer uses the LBFGS algorithm, but does a line search that
    fulfills the Wolff conditions.
    """

    def __init__(self, *args, **kwargs):
        kwargs['use_line_search'] = True
        LBFGS.__init__(self, *args, **kwargs)

#    """Modified version of LBFGS.
#
#    This optimizer uses the LBFGS algorithm, but does a line search for the
#    minimum along the search direction. This is done by issuing an additional
#    force call for each step, thus doubling the number of calculations.
#
#    Additionally the Hessian is reset if the new guess is not sufficiently
#    better than the old one.
#    """
#    def __init__(self, *args, **kwargs):
#        self.dR = kwargs.pop('dR', 0.1)
#        LBFGS.__init__(self, *args, **kwargs)
#
#    def update(self, r, f, r0, f0):
#        """Update everything that is kept in memory
#
#        This function is mostly here to allow for replay_trajectory.
#        """
#        if self.iteration > 0:
#            a1 = abs(np.dot(f.reshape(-1), f0.reshape(-1)))
#            a2 = np.dot(f0.reshape(-1), f0.reshape(-1))
#            if not (a1 <= 0.5 * a2 and a2 != 0):
#                # Reset optimization
#                self.initialize()
#
#        # Note that the reset above will set self.iteration to 0 again
#        # which is why we should check again
#        if self.iteration > 0:
#            s0 = r.reshape(-1) - r0.reshape(-1)
#            self.s.append(s0)
#
#            # We use the gradient which is minus the force!
#            y0 = f0.reshape(-1) - f.reshape(-1)
#            self.y.append(y0)
#
#            rho0 = 1.0 / np.dot(y0, s0)
#            self.rho.append(rho0)
#
#        if self.iteration > self.memory:
#            self.s.pop(0)
#            self.y.pop(0)
#            self.rho.pop(0)
#
#    def determine_step(self, dr):
#        f = self.atoms.get_forces()
#
#        # Unit-vector along the search direction
#        du = dr / np.sqrt(np.dot(dr.reshape(-1), dr.reshape(-1)))
#
#        # We keep the old step determination before we figure
#        # out what is the best to do.
#        maxstep = self.maxstep * np.sqrt(3 * len(self.atoms))
#
#        # Finite difference step using temporary point
#        self.atoms.positions += (du * self.dR)
#        # Decide how much to move along the line du
#        Fp1 = np.dot(f.reshape(-1), du.reshape(-1))
#        Fp2 = np.dot(self.atoms.get_forces().reshape(-1), du.reshape(-1))
#        CR = (Fp1 - Fp2) / self.dR
#        #RdR = Fp1*0.1
#        if CR < 0.0:
#            #print "negcurve"
#            RdR = maxstep
#            #if(abs(RdR) > maxstep):
#            #    RdR = self.sign(RdR) * maxstep
#        else:
#            Fp = (Fp1 + Fp2) * 0.5
#            RdR = Fp / CR
#            if abs(RdR) > maxstep:
#                RdR = np.sign(RdR) * maxstep
#            else:
#                RdR += self.dR * 0.5
#        return du * RdR



import string

import numpy as np

from ase.data import chemical_symbols
from ase.io import string2index
from ase.io.formats import parse_filename

# default fields


def field_specs_on_conditions(calculator_outputs, rank_order):
    if calculator_outputs:
        field_specs = ['i:0', 'el', 'd', 'rd', 'df', 'rdf']
    else:
        field_specs = ['i:0', 'el', 'dx', 'dy', 'dz', 'd', 'rd']
    if rank_order is not None:
        field_specs[0] = 'i:1'
        if rank_order in field_specs:
            for c, i in enumerate(field_specs):
                if i == rank_order:
                    field_specs[c] = i + ':0:1'
        else:
            field_specs.append(rank_order + ':0:1')
    else:
        field_specs[0] = field_specs[0] + ':1'
    return field_specs


def summary_functions_on_conditions(has_calc):
    if has_calc:
        return [rmsd, energy_delta]
    return [rmsd]


def header_alias(h):
    """Replace keyboard characters with Unicode symbols
    for pretty printing"""
    if h == 'i':
        h = 'index'
    elif h == 'an':
        h = 'atomic #'
    elif h == 't':
        h = 'tag'
    elif h == 'el':
        h = 'element'
    elif h[0] == 'd':
        h = h.replace('d', 'Δ')
    elif h[0] == 'r':
        h = 'rank ' + header_alias(h[1:])
    elif h[0] == 'a':
        h = h.replace('a', '<')
        h += '>'
    return h


def prec_round(a, prec=2):
    """
    To make hierarchical sorting different from non-hierarchical sorting
    with floats.
    """
    if a == 0:
        return a
    else:
        s = 1 if a > 0 else -1
        m = np.log10(s * a) // 1
        c = np.log10(s * a) % 1
    return s * np.round(10**c, prec) * 10**m


prec_round = np.vectorize(prec_round)

# end most settings

# this will sort alphabetically by chemical symbol
num2sym = dict(zip(np.argsort(chemical_symbols), chemical_symbols))
# to sort by atomic number, uncomment below
# num2sym = dict(zip(range(len(chemical_symbols)), chemical_symbols))
sym2num = {v: k for k, v in num2sym.items()}

atoms_props = [
    'dx',
    'dy',
    'dz',
    'd',
    't',
    'an',
    'i',
    'el',
    'p1',
    'p2',
    'p1x',
    'p1y',
    'p1z',
    'p2x',
    'p2y',
    'p2z']


def get_field_data(atoms1, atoms2, field):
    if field[0] == 'r':
        field = field[1:]
        rank_order = True
    else:
        rank_order = False

    if field in atoms_props:
        if field == 't':
            data = atoms1.get_tags()
        elif field == 'an':
            data = atoms1.numbers
        elif field == 'el':
            data = np.array([sym2num[sym] for sym in atoms1.symbols])
        elif field == 'i':
            data = np.arange(len(atoms1))
        else:
            if field.startswith('d'):
                y = atoms2.positions - atoms1.positions
            elif field.startswith('p'):
                if field[1] == '1':
                    y = atoms1.positions
                else:
                    y = atoms2.positions

            if field.endswith('x'):
                data = y[:, 0]
            elif field.endswith('y'):
                data = y[:, 1]
            elif field.endswith('z'):
                data = y[:, 2]
            else:
                data = np.linalg.norm(y, axis=1)
    else:
        if field[0] == 'd':
            y = atoms2.get_forces() - atoms1.get_forces()
        elif field[0] == 'a':
            y = (atoms2.get_forces() + atoms1.get_forces()) / 2
        else:
            if field[1] == '1':
                y = atoms1.get_forces()
            else:
                y = atoms2.get_forces()

        if field.endswith('x'):
            data = y[:, 0]
        elif field.endswith('y'):
            data = y[:, 1]
        elif field.endswith('z'):
            data = y[:, 2]
        else:
            data = np.linalg.norm(y, axis=1)

    if rank_order:
        return np.argsort(np.argsort(-data))

    return data


# Summary Functions

def rmsd(atoms1, atoms2):
    dpositions = atoms2.positions - atoms1.positions
    return 'RMSD={:+.1E}'.format(
        np.sqrt((np.linalg.norm(dpositions, axis=1)**2).mean()))


def energy_delta(atoms1, atoms2):
    E1 = atoms1.get_potential_energy()
    E2 = atoms2.get_potential_energy()
    return f'E1 = {E1:+.1E}, E2 = {E2:+.1E}, dE = {E2 - E1:+1.1E}'


def parse_field_specs(field_specs):
    fields = []
    hier = []
    scent = []
    for fs in field_specs:
        fhs = fs.split(':')
        if len(fhs) == 3:
            scent.append(int(fhs[2]))
            hier.append(int(fhs[1]))
            fields.append(fhs[0])
        elif len(fhs) == 2:
            scent.append(-1)
            hier.append(int(fhs[1]))
            fields.append(fhs[0])
        elif len(fhs) == 1:
            scent.append(-1)
            hier.append(-1)
            fields.append(fhs[0])
    mxm = max(hier)
    for c in range(len(hier)):
        if hier[c] < 0:
            mxm += 1
            hier[c] = mxm
    # reversed by convention of numpy lexsort
    hier = np.argsort(hier)[::-1]
    return fields, hier, np.array(scent)

# Class definitions


class MapFormatter(string.Formatter):
    """String formatting method to map string
    mapped to float data field
    used for sorting back to string."""

    def format_field(self, value, spec):
        if spec.endswith('h'):
            value = num2sym[int(value)]
            spec = spec[:-1] + 's'
        return super().format_field(value, spec)


class TableFormat:
    def __init__(self,
                 columnwidth=9,
                 precision=2,
                 representation='E',
                 toprule='=',
                 midrule='-',
                 bottomrule='='):

        self.precision = precision
        self.representation = representation
        self.columnwidth = columnwidth
        self.formatter = MapFormatter().format
        self.toprule = toprule
        self.midrule = midrule
        self.bottomrule = bottomrule

        self.fmt_class = {
            'signed float': "{{: ^{}.{}{}}}".format(
                self.columnwidth,
                self.precision - 1,
                self.representation),
            'unsigned float': "{{:^{}.{}{}}}".format(
                self.columnwidth,
                self.precision - 1,
                self.representation),
            'int': "{{:^{}n}}".format(
                self.columnwidth),
            'str': "{{:^{}s}}".format(
                self.columnwidth),
            'conv': "{{:^{}h}}".format(
                self.columnwidth)}
        fmt = {}
        signed_floats = [
            'dx',
            'dy',
            'dz',
            'dfx',
            'dfy',
            'dfz',
            'afx',
            'afy',
            'afz',
            'p1x',
            'p2x',
            'p1y',
            'p2y',
            'p1z',
            'p2z',
            'f1x',
            'f2x',
            'f1y',
            'f2y',
            'f1z',
            'f2z']
        for sf in signed_floats:
            fmt[sf] = self.fmt_class['signed float']
        unsigned_floats = ['d', 'df', 'af', 'p1', 'p2', 'f1', 'f2']
        for usf in unsigned_floats:
            fmt[usf] = self.fmt_class['unsigned float']
        integers = ['i', 'an', 't'] + ['r' + sf for sf in signed_floats] + \
            ['r' + usf for usf in unsigned_floats]
        for i in integers:
            fmt[i] = self.fmt_class['int']
        fmt['el'] = self.fmt_class['conv']

        self.fmt = fmt


class Table:
    def __init__(self,
                 field_specs,
                 summary_functions=[],
                 tableformat=None,
                 max_lines=None,
                 title='',
                 tablewidth=None):

        self.max_lines = max_lines
        self.summary_functions = summary_functions
        self.field_specs = field_specs

        self.fields, self.hier, self.scent = parse_field_specs(self.field_specs)
        self.nfields = len(self.fields)

        # formatting
        if tableformat is None:
            self.tableformat = TableFormat()
        else:
            self.tableformat = tableformat

        if tablewidth is None:
            self.tablewidth = self.tableformat.columnwidth * self.nfields
        else:
            self.tablewidth = tablewidth

        self.title = title

    def make(self, atoms1, atoms2, csv=False):
        header = self.make_header(csv=csv)
        body = self.make_body(atoms1, atoms2, csv=csv)
        if self.max_lines is not None:
            body = body[:self.max_lines]
        summary = self.make_summary(atoms1, atoms2)

        return '\n'.join([self.title,
                          self.tableformat.toprule * self.tablewidth,
                          header,
                          self.tableformat.midrule * self.tablewidth,
                          body,
                          self.tableformat.bottomrule * self.tablewidth,
                          summary])

    def make_header(self, csv=False):
        if csv:
            return ','.join([header_alias(field) for field in self.fields])

        fields = self.tableformat.fmt_class['str'] * self.nfields
        headers = [header_alias(field) for field in self.fields]

        return self.tableformat.formatter(fields, *headers)

    def make_summary(self, atoms1, atoms2):
        return '\n'.join([summary_function(atoms1, atoms2)
                          for summary_function in self.summary_functions])

    def make_body(self, atoms1, atoms2, csv=False):
        field_data = np.array([get_field_data(atoms1, atoms2, field)
                               for field in self.fields])

        sorting_array = field_data * self.scent[:, np.newaxis]
        sorting_array = sorting_array[self.hier]
        sorting_array = prec_round(sorting_array, self.tableformat.precision)

        field_data = field_data[:, np.lexsort(sorting_array)].transpose()

        if csv:
            rowformat = ','.join(
                ['{:h}' if field == 'el' else '{{:.{}E}}'.format(
                    self.tableformat.precision) for field in self.fields])
        else:
            rowformat = ''.join([self.tableformat.fmt[field]
                                 for field in self.fields])
        body = [
            self.tableformat.formatter(
                rowformat,
                *row) for row in field_data]
        return '\n'.join(body)


default_index = string2index(':')


def slice_split(filename):
    if '@' in filename:
        filename, index = parse_filename(filename, None)
    else:
        filename, index = parse_filename(filename, default_index)
    return filename, index



from typing import Any, Dict, List

import pytest

from ase.utils import (
    deprecated,
    devnull,
    get_python_package_path_description,
    string2index,
    tokenize_version,
)


class DummyWarning(UserWarning):
    pass


def _add(
    a: int = 0, b: int = 0, *args, **kwargs
) -> int:
    return a + b + len(args)


DEPRECATION_MESSAGE = "Test"


@pytest.mark.filterwarnings(
    f"ignore:{DEPRECATION_MESSAGE}:ase.test.test_util.DummyWarning"
)
class TestDeprecatedDecorator:
    @staticmethod
    def test_should_raise_future_warning_by_default() -> None:
        deprecated_add = deprecated(DEPRECATION_MESSAGE)(_add)
        with pytest.warns(FutureWarning, match=DEPRECATION_MESSAGE):
            _ = deprecated_add()

    @staticmethod
    def test_should_satisfy_condition_and_raise_warning_by_default() -> None:
        deprecated_add = deprecated(
            DEPRECATION_MESSAGE, category=DummyWarning
        )(_add)
        with pytest.warns(DummyWarning, match=DEPRECATION_MESSAGE):
            _ = deprecated_add()

    @staticmethod
    def test_should_raise_warning_when_callback_returns_true() -> None:
        def callback(_: List, kwargs: Dict) -> bool:
            return "test" in kwargs

        with pytest.warns(DummyWarning, match=DEPRECATION_MESSAGE):
            deprecated_add = deprecated(
                DEPRECATION_MESSAGE,
                category=DummyWarning,
                callback=callback
            )(_add)
            _ = deprecated_add(test=True)

    @staticmethod
    def test_should_not_raise_warning_when_callback_returns_false() -> None:
        deprecated_add = deprecated(
            DEPRECATION_MESSAGE,
            callback=lambda args, kwargs: False
        )(_add)
        _ = deprecated_add()

    @staticmethod
    def test_should_call_function_correctly() -> None:
        deprecated_add = deprecated(
            DEPRECATION_MESSAGE,
            category=DummyWarning
        )(_add)
        assert deprecated_add(2, 2) == 4

    def test_should_call_callback(self) -> None:
        self.callback_called = False

        def callback(_: List, __: Dict) -> bool:
            self.callback_called = True
            return True

        deprecated_add = deprecated(
            DEPRECATION_MESSAGE,
            category=DummyWarning,
            callback=callback
        )(_add)
        _ = deprecated_add()
        assert self.callback_called

    @staticmethod
    def test_should_call_function_with_modified_args() -> None:
        def double_summands(args: List[int], _):
            for i, val in enumerate(args):
                args[i] = 2 * val

        deprecated_add_double = deprecated(
            DEPRECATION_MESSAGE,
            category=DummyWarning,
            callback=double_summands)(_add)
        assert deprecated_add_double(2, 2) == 8

    @staticmethod
    def test_should_call_function_with_modified_kwargs() -> None:
        def double_summands(_: List[int], kwargs: Dict[str, int]):
            for kwarg, val in kwargs.items():
                kwargs[kwarg] = 2 * val

        deprecated_add_double = deprecated(
            DEPRECATION_MESSAGE,
            category=DummyWarning,
            callback=double_summands)(_add)
        assert deprecated_add_double(a=2, b=2) == 8

    @staticmethod
    def test_should_raise_warning_and_modify_args_if_callback_returns_true(
    ) -> None:
        def limit_args_to_two(args: List, _: Dict[str, Any]):
            del args[2:]

        deprecated_add = deprecated(
            DEPRECATION_MESSAGE,
            category=DummyWarning,
            callback=limit_args_to_two)(_add)
        assert deprecated_add(1, 1, 1) == 2

    @staticmethod
    def test_should_work_when_warning_passed_as_message() -> None:
        deprecated_add = deprecated(FutureWarning(DEPRECATION_MESSAGE))(_add)
        with pytest.warns(FutureWarning, match=DEPRECATION_MESSAGE):
            _ = deprecated_add(2, 2)


def test_deprecated_devnull():
    with pytest.warns(DeprecationWarning):
        devnull.tell()


class TestDeprecationFunctional:
    pass


@pytest.mark.parametrize('v1, v2', [
    ('1', '2'),
    ('a', 'b'),
    ('9.0', '10.0'),
    ('3.8.0', '3.8.1'),
    ('3a', '3b'),
    ('3', '3a'),
])
def test_tokenize_version_lessthan(v1, v2):
    v1 = tokenize_version(v1)
    v2 = tokenize_version(v2)
    assert v1 < v2


def test_tokenize_version_equal():
    version = '3.8x.xx'
    assert tokenize_version(version) == tokenize_version(version)


class DummyIterator:
    def __iter__(self):
        yield from ["test", "bla"]


class Dummy:
    @property
    def __path__(self):
        return DummyIterator()


class TestString2Index:
    """Test `string2index`"""

    def test_zero(self):
        """Test 0"""
        assert string2index("0") == 0

    def test_last(self):
        """Test -1"""
        assert string2index("-1") == -1

    def test_last_two(self):
        """Test -2:"""
        assert string2index("-2:") == slice(-2, None)

    def test_all(self):
        """Test :"""
        assert string2index(":") == slice(None)

    def test_even(self):
        """Test ::2"""
        assert string2index("::2") == slice(None, None, 2)

    def test_three_colons(self):
        """Test invalid index string with three colons"""
        with pytest.raises(TypeError):
            string2index(":::")


def test_get_python_package_path_description():
    assert isinstance(get_python_package_path_description(Dummy()), str)
    # test object not containing __path__
    assert isinstance(get_python_package_path_description(object()), str)



from ase.cell import Cell
from ase.geometry.cell import (
    cell_to_cellpar,
    cellpar_to_cell,
    complete_cell,
    is_orthorhombic,
    orthorhombic,
)
from ase.geometry.distance import distance
from ase.geometry.geometry import (
    conditional_find_mic,
    find_mic,
    get_angles,
    get_angles_derivatives,
    get_dihedrals,
    get_dihedrals_derivatives,
    get_distances,
    get_distances_derivatives,
    get_duplicate_atoms,
    get_layers,
    permute_axes,
    wrap_positions,
)
from ase.geometry.minkowski_reduction import (
    is_minkowski_reduced,
    minkowski_reduce,
)

__all__ = ['Cell', 'wrap_positions', 'complete_cell',
           'is_orthorhombic', 'orthorhombic',
           'get_layers', 'find_mic', 'get_duplicate_atoms',
           'cell_to_cellpar', 'cellpar_to_cell', 'distance',
           'get_angles', 'get_distances', 'get_dihedrals',
           'get_angles_derivatives', 'get_distances_derivatives',
           'get_dihedrals_derivatives', 'conditional_find_mic',
           'permute_axes', 'minkowski_reduce', 'is_minkowski_reduced']



import numpy as np

from ase.units import Bohr, Hartree


class Excitation:
    """Base class for a single excitation"""

    def __init__(self, energy, index, mur, muv=None, magn=None):
        """
        Parameters
        ----------
        energy: float
          Energy realtive to the ground state
        index: int
          Excited state index
        mur: list of three floats or complex numbers
          Length form dipole matrix element
        muv: list of three floats or complex numbers or None
          Velocity form dipole matrix element, default None
        magn: list of three floats or complex numbers or None
          Magnetic matrix element, default None
        """
        self.energy = energy
        self.index = index
        self.mur = mur
        self.muv = muv
        self.magn = magn
        self.fij = 1.

    def outstring(self):
        """Format yourself as a string"""
        string = f'{self.energy:g}  {self.index}  '

        def format_me(me):
            string = ''
            if me.dtype == float:
                for m in me:
                    string += f' {m:g}'
            else:
                for m in me:
                    string += ' {0.real:g}{0.imag:+g}j'.format(m)
            return string

        string += '  ' + format_me(self.mur)
        if self.muv is not None:
            string += '  ' + format_me(self.muv)
        if self.magn is not None:
            string += '  ' + format_me(self.magn)
        string += '\n'

        return string

    @classmethod
    def fromstring(cls, string):
        """Initialize yourself from a string"""
        l = string.split()
        energy = float(l.pop(0))
        index = int(l.pop(0))
        mur = np.array([float(l.pop(0)) for _ in range(3)])
        try:
            muv = np.array([float(l.pop(0)) for _ in range(3)])
        except IndexError:
            muv = None
        try:
            magn = np.array([float(l.pop(0)) for _ in range(3)])
        except IndexError:
            magn = None

        return cls(energy, index, mur, muv, magn)

    def get_dipole_me(self, form='r'):
        """Return the excitations dipole matrix element
        including the occupation factor sqrt(fij)"""
        if form == 'r':  # length form
            me = - self.mur
        elif form == 'v':  # velocity form
            me = - self.muv
        else:
            raise RuntimeError('Unknown form >' + form + '<')
        return np.sqrt(self.fij) * me

    def get_dipole_tensor(self, form='r'):
        """Return the oscillator strength tensor"""
        me = self.get_dipole_me(form)
        return 2 * np.outer(me, me.conj()) * self.energy / Hartree

    def get_oscillator_strength(self, form='r'):
        """Return the excitations dipole oscillator strength."""
        me2_c = self.get_dipole_tensor(form).diagonal().real
        return np.array([np.sum(me2_c) / 3.] + me2_c.tolist())


class ExcitationList(list):
    """Base class for excitions from the ground state"""

    def __init__(self):
        # initialise empty list
        super().__init__()

        # set default energy scale to get eV
        self.energy_to_eV_scale = 1.


def polarizability(exlist, omega, form='v',
                   tensor=False, index=0):
    """Evaluate the photon energy dependent polarizability
    from the sum over states

    Parameters
    ----------
    exlist: ExcitationList
    omega:
        Photon energy (eV)
    form: {'v', 'r'}
        Form of the dipole matrix element, default 'v'
    index: {0, 1, 2, 3}
        0: averaged, 1,2,3:alpha_xx, alpha_yy, alpha_zz, default 0
    tensor: boolean
        if True returns alpha_ij, i,j=x,y,z
        index is ignored, default False

    Returns
    -------
    alpha:
        Unit (e^2 Angstrom^2 / eV).
        Multiply with Bohr * Ha to get (Angstrom^3)
        shape = (omega.shape,) if tensor == False
        shape = (omega.shape, 3, 3) else
    """
    omega = np.asarray(omega)
    om2 = 1. * omega**2
    esc = exlist.energy_to_eV_scale

    if tensor:
        if not np.isscalar(om2):
            om2 = om2[:, None, None]
        alpha = np.zeros(omega.shape + (3, 3),
                         dtype=om2.dtype)
        for ex in exlist:
            alpha += ex.get_dipole_tensor(form=form) / (
                (ex.energy * esc)**2 - om2)
    else:
        alpha = np.zeros_like(om2)
        for ex in exlist:
            alpha += ex.get_oscillator_strength(form=form)[index] / (
                (ex.energy * esc)**2 - om2)

    return alpha * Bohr**2 * Hartree



"""Build crystalline systems"""
from math import sqrt
from typing import Any

from ase.atoms import Atoms
from ase.data import atomic_numbers, chemical_symbols, reference_states
from ase.symbols import string2symbols
from ase.utils import plural


def incompatible_cell(*, want, have):
    return RuntimeError(f'Cannot create {want} cell for {have} structure')


def bulk(
    name: str,
    crystalstructure: str = None,
    a: float = None,
    b: float = None,
    c: float = None,
    *,
    alpha: float = None,
    covera: float = None,
    u: float = None,
    orthorhombic: bool = False,
    cubic: bool = False,
    basis=None,
) -> Atoms:
    """Creating bulk systems.

    Crystal structure and lattice constant(s) will be guessed if not
    provided.

    name: str
        Chemical symbol or symbols as in 'MgO' or 'NaCl'.
    crystalstructure: str
        Must be one of sc, fcc, bcc, tetragonal, bct, hcp, rhombohedral,
        orthorhombic, mcl, diamond, zincblende, rocksalt, cesiumchloride,
        fluorite or wurtzite.
    a: float
        Lattice constant.
    b: float
        Lattice constant.  If only a and b is given, b will be interpreted
        as c instead.
    c: float
        Lattice constant.
    alpha: float
        Angle in degrees for rhombohedral lattice.
    covera: float
        c/a ratio used for hcp.  Default is ideal ratio: sqrt(8/3).
    u: float
        Internal coordinate for Wurtzite structure.
    orthorhombic: bool
        Construct orthorhombic unit cell instead of primitive cell
        which is the default.
    cubic: bool
        Construct cubic unit cell if possible.
    """

    if c is None and b is not None:
        # If user passes (a, b) positionally, we want it as (a, c) instead:
        c, b = b, c

    if covera is not None and c is not None:
        raise ValueError("Don't specify both c and c/a!")

    xref = ''
    ref: Any = {}

    if name in chemical_symbols:  # single element
        atomic_number = atomic_numbers[name]
        ref = reference_states[atomic_number]
        if ref is None:
            ref = {}  # easier to 'get' things from empty dictionary than None
        else:
            xref = ref['symmetry']

        if crystalstructure is None:
            # `ref` requires `basis` but not given and not pre-defined
            if basis is None and 'basis' in ref and ref['basis'] is None:
                raise ValueError('This structure requires an atomic basis')
            if xref == 'cubic':
                # P and Mn are listed as 'cubic' but the lattice constants
                # are 7 and 9.  They must be something other than simple cubic
                # then. We used to just return the cubic one but that must
                # have been wrong somehow.  --askhl
                raise ValueError(
                    f'The reference structure of {name} is not implemented')

    # Mapping of name to number of atoms in primitive cell.
    structures = {'sc': 1, 'fcc': 1, 'bcc': 1,
                  'tetragonal': 1,
                  'bct': 1,
                  'hcp': 1,
                  'rhombohedral': 1,
                  'orthorhombic': 1,
                  'mcl': 1,
                  'diamond': 1,
                  'zincblende': 2, 'rocksalt': 2, 'cesiumchloride': 2,
                  'fluorite': 3, 'wurtzite': 2}

    if crystalstructure is None:
        crystalstructure = xref
        if crystalstructure not in structures:
            raise ValueError(f'No suitable reference data for bulk {name}.'
                             f'  Reference data: {ref}')

    magmom_per_atom = None
    if crystalstructure == xref:
        magmom_per_atom = ref.get('magmom_per_atom')

    if crystalstructure not in structures:
        raise ValueError(f'Unknown structure: {crystalstructure}.')

    # Check name:
    natoms = len(string2symbols(name))
    natoms0 = structures[crystalstructure]
    if natoms != natoms0:
        raise ValueError('Please specify {} for {} and not {}'
                         .format(plural(natoms0, 'atom'),
                                 crystalstructure, natoms))

    if alpha is None:
        alpha = ref.get('alpha')

    if a is None:
        if xref != crystalstructure:
            raise ValueError('You need to specify the lattice constant.')
        if 'a' in ref:
            a = ref['a']
        else:
            raise KeyError(f'No reference lattice parameter "a" for "{name}"')

    if b is None:
        bovera = ref.get('b/a')
        if bovera is not None and a is not None:
            b = bovera * a

    if crystalstructure in ['hcp', 'wurtzite']:
        if c is not None:
            covera = c / a
        elif covera is None:
            if xref == crystalstructure:
                covera = ref['c/a']
            else:
                covera = sqrt(8 / 3)

    if covera is None:
        covera = ref.get('c/a')
        if c is None and covera is not None:
            c = covera * a

    if crystalstructure == 'bct':
        from ase.lattice import BCT
        if basis is None:
            basis = ref.get('basis')
        if basis is not None:
            natoms = len(basis)
        lat = BCT(a=a, c=c)
        atoms = Atoms([name] * natoms, cell=lat.tocell(), pbc=True,
                      scaled_positions=basis)
    elif crystalstructure == 'rhombohedral':
        atoms = _build_rhl(name, a, alpha, basis)
    elif crystalstructure == 'orthorhombic':
        atoms = Atoms(name, cell=[a, b, c], pbc=True)
    elif orthorhombic:
        atoms = _orthorhombic_bulk(name, crystalstructure, a, covera, u)
    elif cubic:
        atoms = _cubic_bulk(name, crystalstructure, a)
    else:
        atoms = _primitive_bulk(name, crystalstructure, a, covera, u)

    if magmom_per_atom is not None:
        magmoms = [magmom_per_atom] * len(atoms)
        atoms.set_initial_magnetic_moments(magmoms)

    if cubic or orthorhombic:
        assert atoms.cell.orthorhombic

    return atoms


def _build_rhl(name, a, alpha, basis):
    from ase.lattice import RHL
    lat = RHL(a, alpha)
    cell = lat.tocell()
    if basis is None:
        # RHL: Given by A&M as scaled coordinates "x" of cell.sum(0):
        basis_x = reference_states[atomic_numbers[name]]['basis_x']
        basis = basis_x[:, None].repeat(3, axis=1)
    natoms = len(basis)
    return Atoms([name] * natoms, cell=cell, scaled_positions=basis, pbc=True)


def _orthorhombic_bulk(name, crystalstructure, a, covera=None, u=None):
    if crystalstructure in ('sc', 'bcc', 'cesiumchloride'):
        atoms = _cubic_bulk(name, crystalstructure, a)
    elif crystalstructure == 'fcc':
        b = a / sqrt(2)
        cell = (b, b, a)
        scaled_positions = ((0.0, 0.0, 0.0), (0.5, 0.5, 0.5))
        atoms = Atoms(2 * name, cell=cell, scaled_positions=scaled_positions)
    elif crystalstructure == 'hcp':
        cell = (a, a * sqrt(3), covera * a)
        scaled_positions = [
            (0.0, 0 / 6, 0.0),
            (0.5, 3 / 6, 0.0),
            (0.5, 1 / 6, 0.5),
            (0.0, 4 / 6, 0.5),
        ]
        atoms = Atoms(4 * name, cell=cell, scaled_positions=scaled_positions)
    elif crystalstructure == 'diamond':
        b = a / sqrt(2)
        cell = (b, b, a)
        scaled_positions = [
            (0.0, 0.0, 0.0), (0.5, 0.0, 0.25),
            (0.5, 0.5, 0.5), (0.0, 0.5, 0.75),
        ]
        atoms = Atoms(4 * name, cell=cell, scaled_positions=scaled_positions)
    elif crystalstructure == 'rocksalt':
        b = a / sqrt(2)
        cell = (b, b, a)
        scaled_positions = [
            (0.0, 0.0, 0.0), (0.5, 0.5, 0.0),
            (0.5, 0.5, 0.5), (0.0, 0.0, 0.5),
        ]
        atoms = Atoms(2 * name, cell=cell, scaled_positions=scaled_positions)
    elif crystalstructure == 'zincblende':
        symbol0, symbol1 = string2symbols(name)
        atoms = _orthorhombic_bulk(symbol0, 'diamond', a)
        atoms.symbols[[1, 3]] = symbol1
    elif crystalstructure == 'wurtzite':
        cell = (a, a * sqrt(3), covera * a)
        u = u or 0.25 + 1 / 3 / covera**2
        scaled_positions = [
            (0.0, 0 / 6, 0.0), (0.0, 2 / 6, 0.5 - u),
            (0.0, 2 / 6, 0.5), (0.0, 0 / 6, 1.0 - u),
            (0.5, 3 / 6, 0.0), (0.5, 5 / 6, 0.5 - u),
            (0.5, 5 / 6, 0.5), (0.5, 3 / 6, 1.0 - u),
        ]
        atoms = Atoms(4 * name, cell=cell, scaled_positions=scaled_positions)
    else:
        raise incompatible_cell(want='orthorhombic', have=crystalstructure)

    atoms.pbc = True

    return atoms


def _cubic_bulk(name: str, crystalstructure: str, a: float) -> Atoms:
    cell = (a, a, a)
    if crystalstructure == 'sc':
        atoms = Atoms(name, cell=cell)
    elif crystalstructure == 'fcc':
        scaled_positions = [
            (0.0, 0.0, 0.0),
            (0.0, 0.5, 0.5),
            (0.5, 0.0, 0.5),
            (0.5, 0.5, 0.0),
        ]
        atoms = Atoms(4 * name, cell=cell, scaled_positions=scaled_positions)
    elif crystalstructure == 'bcc':
        scaled_positions = [
            (0.0, 0.0, 0.0),
            (0.5, 0.5, 0.5),
        ]
        atoms = Atoms(2 * name, cell=cell, scaled_positions=scaled_positions)
    elif crystalstructure == 'diamond':
        scaled_positions = [
            (0.0, 0.0, 0.0), (0.25, 0.25, 0.25),
            (0.0, 0.5, 0.5), (0.25, 0.75, 0.75),
            (0.5, 0.0, 0.5), (0.75, 0.25, 0.75),
            (0.5, 0.5, 0.0), (0.75, 0.75, 0.25),
        ]
        atoms = Atoms(8 * name, cell=cell, scaled_positions=scaled_positions)
    elif crystalstructure == 'cesiumchloride':
        symbol0, symbol1 = string2symbols(name)
        atoms = _cubic_bulk(symbol0, 'bcc', a)
        atoms.symbols[[1]] = symbol1
    elif crystalstructure == 'zincblende':
        symbol0, symbol1 = string2symbols(name)
        atoms = _cubic_bulk(symbol0, 'diamond', a)
        atoms.symbols[[1, 3, 5, 7]] = symbol1
    elif crystalstructure == 'rocksalt':
        scaled_positions = [
            (0.0, 0.0, 0.0), (0.5, 0.0, 0.0),
            (0.0, 0.5, 0.5), (0.5, 0.5, 0.5),
            (0.5, 0.0, 0.5), (0.0, 0.0, 0.5),
            (0.5, 0.5, 0.0), (0.0, 0.5, 0.0),
        ]
        atoms = Atoms(4 * name, cell=cell, scaled_positions=scaled_positions)
    elif crystalstructure == 'fluorite':
        scaled_positions = [
            (0.00, 0.00, 0.00), (0.25, 0.25, 0.25), (0.75, 0.75, 0.75),
            (0.00, 0.50, 0.50), (0.25, 0.75, 0.75), (0.75, 0.25, 0.25),
            (0.50, 0.00, 0.50), (0.75, 0.25, 0.75), (0.25, 0.75, 0.25),
            (0.50, 0.50, 0.00), (0.75, 0.75, 0.25), (0.25, 0.25, 0.75),
        ]
        atoms = Atoms(4 * name, cell=cell, scaled_positions=scaled_positions)
    else:
        raise incompatible_cell(want='cubic', have=crystalstructure)

    atoms.pbc = True

    return atoms


def _primitive_bulk(name, crystalstructure, a, covera=None, u=None):
    if crystalstructure == 'sc':
        atoms = Atoms(name, cell=(a, a, a))
    elif crystalstructure == 'fcc':
        b = 0.5 * a
        cell = ((0, b, b), (b, 0, b), (b, b, 0))
        atoms = Atoms(name, cell=cell)
    elif crystalstructure == 'bcc':
        b = 0.5 * a
        cell = ((-b, b, b), (b, -b, b), (b, b, -b))
        atoms = Atoms(name, cell=cell)
    elif crystalstructure == 'hcp':
        c = covera * a
        cell = ((a, 0, 0), (-0.5 * a, 0.5 * sqrt(3) * a, 0), (0, 0, c))
        scaled_positions = [
            (0 / 3, 0 / 3, 0.0),
            (1 / 3, 2 / 3, 0.5),
        ]
        atoms = Atoms(2 * name, cell=cell, scaled_positions=scaled_positions)
    elif crystalstructure == 'diamond':
        atoms = \
            _primitive_bulk(name, 'fcc', a) + \
            _primitive_bulk(name, 'fcc', a)
        atoms.positions[1, :] += 0.25 * a
    elif crystalstructure == 'rocksalt':
        symbol0, symbol1 = string2symbols(name)
        atoms = \
            _primitive_bulk(symbol0, 'fcc', a) + \
            _primitive_bulk(symbol1, 'fcc', a)
        atoms.positions[1, 0] += 0.5 * a
    elif crystalstructure == 'cesiumchloride':
        symbol0, symbol1 = string2symbols(name)
        atoms = \
            _primitive_bulk(symbol0, 'sc', a) + \
            _primitive_bulk(symbol1, 'sc', a)
        atoms.positions[1, :] += 0.5 * a
    elif crystalstructure == 'zincblende':
        symbol0, symbol1 = string2symbols(name)
        atoms = \
            _primitive_bulk(symbol0, 'fcc', a) + \
            _primitive_bulk(symbol1, 'fcc', a)
        atoms.positions[1, :] += 0.25 * a
    elif crystalstructure == 'fluorite':
        symbol0, symbol1, symbol2 = string2symbols(name)
        atoms = \
            _primitive_bulk(symbol0, 'fcc', a) + \
            _primitive_bulk(symbol1, 'fcc', a) + \
            _primitive_bulk(symbol2, 'fcc', a)
        atoms.positions[1, :] += 0.25 * a
        atoms.positions[2, :] += 0.75 * a
    elif crystalstructure == 'wurtzite':
        c = covera * a
        cell = ((a, 0, 0), (-0.5 * a, 0.5 * sqrt(3) * a, 0), (0, 0, c))
        u = u or 0.25 + 1 / 3 / covera**2
        scaled_positions = [
            (0 / 3, 0 / 3, 0.0), (1 / 3, 2 / 3, 0.5 - u),
            (1 / 3, 2 / 3, 0.5), (0 / 3, 0 / 3, 1.0 - u),
        ]
        atoms = Atoms(2 * name, cell=cell, scaled_positions=scaled_positions)
    else:
        raise incompatible_cell(want='primitive', have=crystalstructure)

    atoms.pbc = True

    return atoms



from collections.abc import Iterable
from typing import Optional

import numpy as np

from ase import Atoms
from ase.calculators.calculator import BaseCalculator, all_properties


class FiniteDifferenceCalculator(BaseCalculator):
    """Wrapper calculator using the finite-difference method.

    The forces and the stress are computed using the finite-difference method.

    .. versionadded:: 3.24.0
    """

    implemented_properties = all_properties

    def __init__(
        self,
        calc: BaseCalculator,
        eps_disp: float = 1e-6,
        eps_strain: float = 1e-6,
    ) -> None:
        """

        Parameters
        ----------
        calc : :class:`~ase.calculators.calculator.BaseCalculator`
            ASE Calculator object to be wrapped.
        eps_disp : float, default 1e-6
            Displacement used for computing forces.
        eps_strain : float, default 1e-6
            Strain used for computing stress.

        """
        super().__init__()
        self.calc = calc
        self.eps_disp = eps_disp
        self.eps_strain = eps_strain

    def calculate(self, atoms: Atoms, properties, system_changes) -> None:
        atoms = atoms.copy()  # copy to not mess up original `atoms`
        atoms.calc = self.calc
        self.results = {
            'energy': atoms.get_potential_energy(),
            'forces': calculate_numerical_forces(atoms, eps=self.eps_disp),
            'stress': calculate_numerical_stress(atoms, eps=self.eps_strain),
        }
        self.results['free_energy'] = self.results['energy']


def _numeric_force(
    atoms: Atoms,
    iatom: int,
    icart: int,
    eps: float = 1e-6,
) -> float:
    """Calculate numerical force on a specific atom along a specific direction.

    Parameters
    ----------
    atoms : :class:`~ase.Atoms`
        ASE :class:`~ase.Atoms` object.
    iatom : int
        Index of atoms.
    icart : {0, 1, 2}
        Index of Cartesian component.
    eps : float, default 1e-6
        Displacement.

    """
    p0 = atoms.get_positions()
    p = p0.copy()
    p[iatom, icart] = p0[iatom, icart] + eps
    atoms.set_positions(p, apply_constraint=False)
    eplus = atoms.get_potential_energy()
    p[iatom, icart] = p0[iatom, icart] - eps
    atoms.set_positions(p, apply_constraint=False)
    eminus = atoms.get_potential_energy()
    atoms.set_positions(p0, apply_constraint=False)
    return (eminus - eplus) / (2 * eps)


def calculate_numerical_forces(
    atoms: Atoms,
    eps: float = 1e-6,
    iatoms: Optional[Iterable[int]] = None,
    icarts: Optional[Iterable[int]] = None,
) -> np.ndarray:
    """Calculate forces numerically based on the finite-difference method.

    Parameters
    ----------
    atoms : :class:`~ase.Atoms`
        ASE :class:`~ase.Atoms` object.
    eps : float, default 1e-6
        Displacement.
    iatoms : Optional[Iterable[int]]
        Indices of atoms for which forces are computed.
        By default, all atoms are considered.
    icarts : Optional[Iterable[int]]
        Indices of Cartesian coordinates for which forces are computed.
        By default, all three coordinates are considered.

    Returns
    -------
    forces : np.ndarray
        Forces computed numerically based on the finite-difference method.

    """
    if iatoms is None:
        iatoms = range(len(atoms))
    if icarts is None:
        icarts = [0, 1, 2]
    return np.array(
        [
            [_numeric_force(atoms, iatom, icart, eps) for icart in icarts]
            for iatom in iatoms
        ]
    )


def calculate_numerical_stress(
    atoms: Atoms,
    eps: float = 1e-6,
    voigt: bool = True,
) -> np.ndarray:
    """Calculate stress numerically based on the finite-difference method.

    Parameters
    ----------
    atoms : :class:`~ase.Atoms`
        ASE :class:`~ase.Atoms` object.
    eps : float, default 1e-6
        Strain in the Voigt notation.
    voigt : bool, default True
        If True, the stress is returned in the Voigt notation.

    Returns
    -------
    stress : np.ndarray
        Stress computed numerically based on the finite-difference method.

    """
    stress = np.zeros((3, 3), dtype=float)

    cell = atoms.cell.copy()
    volume = atoms.get_volume()
    for i in range(3):
        x = np.eye(3)
        x[i, i] = 1.0 + eps
        atoms.set_cell(cell @ x, scale_atoms=True)
        eplus = atoms.get_potential_energy(force_consistent=True)

        x[i, i] = 1.0 - eps
        atoms.set_cell(cell @ x, scale_atoms=True)
        eminus = atoms.get_potential_energy(force_consistent=True)

        stress[i, i] = (eplus - eminus) / (2 * eps * volume)
        x[i, i] = 1.0

        j = i - 2
        x[i, j] = x[j, i] = +0.5 * eps
        atoms.set_cell(cell @ x, scale_atoms=True)
        eplus = atoms.get_potential_energy(force_consistent=True)

        x[i, j] = x[j, i] = -0.5 * eps
        atoms.set_cell(cell @ x, scale_atoms=True)
        eminus = atoms.get_potential_energy(force_consistent=True)

        stress[i, j] = stress[j, i] = (eplus - eminus) / (2 * eps * volume)

    atoms.set_cell(cell, scale_atoms=True)

    return stress.flat[[0, 4, 8, 5, 2, 1]] if voigt else stress



import numpy as np
import pytest

from ase import Atoms
from ase.build import molecule
from ase.calculators.emt import EMT
from ase.constraints import FixAtoms, FixBondLength
from ase.db import connect
from ase.io import read


def test_db2(testdir, dbtype2, get_db_name):
    dbtype = dbtype2
    name = get_db_name(dbtype)

    c = connect(name)
    print(name, c)

    id = c.reserve(abc=7)
    c.delete([d.id for d in c.select(abc=7)])
    id = c.reserve(abc=7)
    assert c[id].abc == 7

    a = c.get_atoms(id)
    c.write(Atoms())
    ch4 = molecule('CH4', calculator=EMT())
    ch4.constraints = [FixAtoms(indices=[1]),
                       FixBondLength(0, 2)]
    f1 = ch4.get_forces()
    print(f1)

    c.delete([d.id for d in c.select(C=1)])
    chi = np.array([1 + 0.5j, 0.5])
    if 'db' in name:
        kvp = {'external_tables':
               {'blabla': {'a': 1, 'b': 2, 'c': 3},
                'lala': {'a': 0.01, 'b': 0.02, 'c': 0.0}}}

    else:
        kvp = {'a': 1}

    id = c.write(ch4, key_value_pairs=kvp,
                 data={'1-butyne': 'bla-bla', 'chi': chi})

    row = c.get(id)
    print(row.data['1-butyne'], row.data.chi)
    assert (row.data.chi == chi).all(), (row.data.chi, chi)
    print(row)

    assert len(c.get_atoms(C=1).constraints) == 2

    f2 = c.get(C=1).forces
    assert abs(f2.sum(0)).max() < 1e-14
    f3 = c.get_atoms(C=1).get_forces()
    assert abs(f1 - f3).max() < 1e-14

    a = read(name, index=f'id={id}')[0]
    f4 = a.get_forces()
    assert abs(f1 - f4).max() < 1e-14

    with pytest.raises(ValueError):
        c.update(id, abc={'a': 42})

    c.update(id, grr='hmm')
    row = c.get(C=1)
    assert row.id == id
    assert (row.data.chi == chi).all()

    for row in c.select(include_data=False):
        assert len(row.data) == 0

    with pytest.raises(ValueError):
        c.write(ch4, foo=['bar', 2])  # not int, bool, float or str

    with pytest.raises(ValueError):
        c.write(Atoms(), pi='3.14')  # float as a string

    with pytest.raises(ValueError):
        c.write(Atoms(), pi_rounded='3')  # int as a string

    with pytest.raises(ValueError):
        c.write(Atoms(), relaxed='False')  # bool as a string

    with pytest.raises(ValueError):
        c.write(Atoms(), fmax=0.0)  # reserved word

    with pytest.raises(ValueError):
        c.write(Atoms(), S=42)  # chemical symbol as key

    id = c.write(Atoms(),
                 b=np.bool_(True),
                 i=np.int64(42),
                 n=np.nan,
                 x=np.inf,
                 s='NaN2',
                 A=42)
    row = c[id]
    assert isinstance(row.b, bool)
    assert isinstance(row.i, int)
    assert np.isnan(row.n)
    assert np.isinf(row.x)

    # Make sure deleting a single key works:
    id = c.write(Atoms(), key=7)
    c.update(id, delete_keys=['key'])
    assert 'key' not in c[id]

    e = [row.get('energy') for row in c.select(sort='energy')]
    assert len(e) == 5 and abs(e[0] - 1.991) < 0.0005

    # Test the offset keyword
    ids = [row.get('id') for row in c.select()]
    offset = 2
    assert next(c.select(offset=offset)).id == ids[offset]

    assert 'grr' in c.get_all_key_names()


def test_sqlite_user(testdir):
    """Make sure user=someone works.  Is called username in SQLite."""
    db = connect('test_user.db')
    assert list(db.select(user='someone')) == []



import numpy as np
import pytest

from ase import Atoms
from ase.build.tools import reduce_lattice
from ase.calculators.emt import EMT


def test_reduce_lattice():
    """Test that reduce_lattice() correctly in-place-reduces lattice.

    We test that the 60 degree hex cell reduces to 120 degree cell.
    To verify that the full system is physically equivalent, we verify
    that the EMT energy is the same after the transformation."""

    origcellpar = [3, 3, 4, 90, 90, 60]
    ref_cellpar = list(origcellpar)
    ref_cellpar[-1] = 120

    symbols = 'CuAgAuPt'
    scaled_positions = np.random.RandomState(42).random((4, 3))

    atoms = Atoms(symbols=symbols,
                  scaled_positions=scaled_positions,
                  cell=origcellpar, pbc=True)

    reduced_atoms = atoms.copy()
    reduce_lattice(reduced_atoms)

    assert reduced_atoms.cell.cellpar() == pytest.approx(ref_cellpar)
    assert emt_energy(atoms) == pytest.approx(emt_energy(reduced_atoms))


def emt_energy(atoms):
    atoms = atoms.copy()
    atoms.calc = EMT()
    return atoms.get_potential_energy()



import numpy as np
import pytest

from ase import Atoms


@pytest.mark.calculator_lite()
@pytest.mark.calculator("lammpslib")
def test_lammpslib_small_nonperiodic(factory, dimer_params, calc_params_NiH):
    """Test that lammpslib handle nonperiodic cases where the cell size
    in some directions is small (for example for a dimer)"""
    # Make a dimer in a nonperiodic box
    dimer = Atoms(**dimer_params)

    # Set of calculator
    calc = factory.calc(**calc_params_NiH)
    dimer.calc = calc

    # Check energy
    energy_ref = -1.10756669119
    energy = dimer.get_potential_energy()
    print(f"Computed energy: {energy}")
    assert energy == pytest.approx(energy_ref, rel=1e-4)

    # Check forces
    forces_ref = np.array(
        [[-0.9420162329811532, 0.0, 0.0], [+0.9420162329811532, 0.0, 0.0]]
    )
    forces = dimer.get_forces()
    print("Computed forces:")
    print(np.array2string(forces))
    assert forces == pytest.approx(forces_ref, rel=1e-4)

    energies = dimer.get_potential_energies()
    assert len(energies) == len(dimer)
    assert sum(energies) == pytest.approx(energy_ref, rel=1e-4)
    assert energies[0] == pytest.approx(energies[1], rel=1e-4)



from math import cos, radians, sin

import pytest

from ase import Atoms
from ase.calculators.emt import EMT
from ase.constraints import FixAtoms
from ase.mep import NEB
from ase.optimize import BFGS, QuasiNewton

# http://jcp.aip.org/resource/1/jcpsa6/v97/i10/p7507_s1
doo = 2.74
doht = 0.957
doh = 0.977
angle = radians(104.5)


@pytest.fixture()
def initial():
    return Atoms('HOHOH',
                 positions=[(-sin(angle) * doht, 0., cos(angle) * doht),
                            (0., 0., 0.),
                            (0., 0., doh),
                            (0., 0., doo),
                            (sin(angle) * doht, 0., doo - cos(angle) * doht)])


@pytest.fixture()
def final(initial):
    atoms = initial.copy()
    atoms.positions[2, 2] = doo - doh
    return atoms


def test_emt_h3o2m(initial, final, testdir):
    # Make band:
    images = [initial.copy()]
    for _ in range(3):
        images.append(initial.copy())
    images.append(final.copy())
    neb = NEB(images, climb=True)

    # Set constraints and calculator:
    constraint = FixAtoms(indices=[1, 3])  # fix OO
    for image in images:
        image.calc = EMT()
        image.set_constraint(constraint)

    for image in images:  # O-H(shared) distance
        print(image.get_distance(1, 2), image.get_potential_energy())

    # Relax initial and final states:
    # One would have to optimize more tightly in order to get
    # symmetric anion from both images[0] and [1], but
    # if one optimizes tightly one gets rotated(H2O) ... OH- instead
    dyn1 = QuasiNewton(images[0])
    dyn1.run(fmax=0.01)
    dyn2 = QuasiNewton(images[-1])
    dyn2.run(fmax=0.01)

    # Interpolate positions between initial and final states:
    neb.interpolate()

    for image in images:
        print(image.get_distance(1, 2), image.get_potential_energy())

    with BFGS(neb, trajectory='emt_h3o2m.traj') as dyn:
        dyn.run(fmax=0.05)

    for image in images:
        print(image.get_distance(1, 2), image.get_potential_energy())



import numpy as np

from ase import Atoms
from ase.calculators.counterions import AtomicCounterIon as ACI
from ase.calculators.fd import calculate_numerical_forces


def test_aic():
    """Test Atomic Counter Ion calc forces."""

    atoms = Atoms('2Na', positions=np.array([[0, 0, 0], [0, 0, 4]]))

    atoms.calc = ACI(1, 1.6642, 0.0001201186, rc=4.5)
    f = atoms.get_forces()
    df = calculate_numerical_forces(atoms, 1e-6) - f
    print(df)
    assert abs(df).max() < 2e-6



import pytest

from ase import Atoms
from ase.calculators.gaussian import Gaussian, GaussianIRC, GaussianOptimizer
from ase.io import read
from ase.optimize import LBFGS


@pytest.fixture()
def atoms():
    return Atoms('CHO',
                 [[0.0, 0.0, 0.0],
                  [0.0, 0.0, 1.35],
                  [1.178513, 0.0, -0.416662]],
                 magmoms=[0.5, 0.0, 0.5])


def get_calc(**kwargs):
    kwargs.update(mem='100MW', method='hf', basis='sto-3g')
    return Gaussian(**kwargs)


def test_optimizer(atoms, gaussian_factory):
    pos = atoms.positions.copy()
    atoms.calc = get_calc(label='opt', scf='qc')
    opt_gauss = GaussianOptimizer(atoms)
    opt_gauss.run(fmax='tight')
    e_gaussopt = read('opt.log', index=-1).get_potential_energy()

    atoms.positions[:] = pos
    atoms.calc.set_label('sp')
    with LBFGS(atoms) as opt_ase:
        opt_ase.run(fmax=1e-2)
    e_aseopt = atoms.get_potential_energy()
    assert e_gaussopt - e_aseopt == pytest.approx(0., abs=1e-3)


def test_irc(atoms, gaussian_factory):
    calc_ts = get_calc(label='ts', chk='ts.chk')
    ts = GaussianOptimizer(atoms, calc_ts)
    ts.run(opt='calcall,ts,noeigentest')
    tspos = atoms.positions.copy()

    atoms.calc = get_calc(label='sp', chk='sp.chk', freq='')
    e_ts = atoms.get_potential_energy()

    calc_irc_for = get_calc(label='irc_for', oldchk='sp', chk='irc_for.chk')
    irc_for = GaussianIRC(atoms, calc_irc_for)
    irc_for.run(direction='forward', irc='rcfc')
    e_for = read('irc_for.log', index=-1).get_potential_energy()

    atoms.positions[:] = tspos
    calc_irc_rev = get_calc(label='irc_rev', oldchk='sp', chk='irc_rev.chk')
    irc_rev = GaussianIRC(atoms, calc_irc_rev)
    irc_rev.run(direction='reverse', irc='rcfc')
    e_rev = read('irc_rev.log', index=-1).get_potential_energy()

    assert e_ts - e_for == pytest.approx(1.282, abs=1e-3)
    assert e_ts - e_rev == pytest.approx(0.201, abs=1e-3)



import numpy as np
import pytest

from ase.geometry import (
    get_angles,
    get_angles_derivatives,
    get_dihedrals,
    get_dihedrals_derivatives,
    get_distances_derivatives,
)


def get_numerical_derivatives(positions, mode, epsilon):
    if mode == 'distance':
        mode_n = 0  # get derivative for bond 0-1
    elif mode == 'angle':
        mode_n = 1  # get derivative for angle 0-1-2
    elif mode == 'dihedral':
        mode_n = 2  # get derivative for dihedral 0-1-2-3
    derivs = np.zeros((2 + mode_n, 3))
    for i in range(2 + mode_n):
        for j in range(3):
            pos = positions.copy()
            pos[i, j] -= epsilon
            if mode == 'distance':
                minus = np.linalg.norm(pos[1] - pos[0])
            elif mode == 'angle':
                minus = get_angles(
                    [pos[0] - pos[1]],
                    [pos[2] - pos[1]],
                )[0]  # size-1 array -> scalar
            elif mode == 'dihedral':
                minus = get_dihedrals(
                    [pos[1] - pos[0]],
                    [pos[2] - pos[1]],
                    [pos[3] - pos[2]],
                )[0]  # size-1 array -> scalar
            pos[i, j] += 2 * epsilon
            if mode == 'distance':
                plus = np.linalg.norm(pos[1] - pos[0])
            elif mode == 'angle':
                plus = get_angles(
                    [pos[0] - pos[1]],
                    [pos[2] - pos[1]],
                )[0]  # size-1 array -> scalar
            elif mode == 'dihedral':
                plus = get_dihedrals(
                    [pos[1] - pos[0]],
                    [pos[2] - pos[1]],
                    [pos[3] - pos[2]],
                )[0]  # size-1 array -> scalar
            derivs[i, j] = (plus - minus) / (2 * epsilon)
    return derivs


def _get_positions():
    # example: positions for H2O2 molecule
    return np.asarray([
        [0.840, 0.881, 0.237],    # H
        [0., 0.734, -0.237],      # O
        [0., -0.734, -0.237],     # O
        [-0.840, -0.881, 0.237],  # H
    ])


def test_atoms_angle():
    pos = _get_positions()

    # analytical derivatives in Angstrom/Angstrom, i.e. degrees/Angstrom
    distances_derivs = get_distances_derivatives([pos[1] - pos[0]])[0]
    angles_derivs = get_angles_derivatives([pos[0] - pos[1]],
                                           [pos[2] - pos[1]])[0]
    dihedrals_derivs = get_dihedrals_derivatives([pos[1] - pos[0]],
                                                 [pos[2] - pos[1]],
                                                 [pos[3] - pos[2]])[0]

    # numerical approximations to derivatives using finite differences
    epsilon = 1e-5
    num_distances_derivs = get_numerical_derivatives(pos, mode='distance',
                                                     epsilon=epsilon)
    num_angles_derivs = get_numerical_derivatives(pos, mode='angle',
                                                  epsilon=epsilon)
    num_dihedrals_derivs = get_numerical_derivatives(pos, mode='dihedral',
                                                     epsilon=epsilon)

    # print(distances_derivs - num_distances_derivs)
    # print(angles_derivs - num_angles_derivs)
    # print(dihedrals_derivs - num_dihedrals_derivs)

    # finite differences versus analytical results
    assert num_distances_derivs == pytest.approx(distances_derivs, abs=1e-8)
    assert num_angles_derivs == pytest.approx(angles_derivs, abs=1e-8)
    assert num_dihedrals_derivs == pytest.approx(dihedrals_derivs, abs=1e-8)

    # derivatives of multiple internal coordinates at once
    assert (distances_derivs ==
            get_distances_derivatives([pos[1] - pos[0],
                                       pos[1] - pos[0]])[0]).all()
    assert (angles_derivs ==
            get_angles_derivatives([pos[0] - pos[1], pos[0] - pos[1]],
                                   [pos[2] - pos[1],
                                    pos[2] - pos[1]])[0]).all()
    assert (dihedrals_derivs ==
            get_dihedrals_derivatives([pos[1] - pos[0], pos[1] - pos[0]],
                                      [pos[2] - pos[1], pos[2] - pos[1]],
                                      [pos[3] - pos[2],
                                       pos[3] - pos[2]])[0]).all()


class TestNondestructive:
    """Test that plural derivative functions do not mutate inputs (#1560)."""
    @staticmethod
    def test_distances_derivatives():
        """Test `get_distances_derivatives`."""
        pos = _get_positions()
        v01s = np.atleast_2d(pos[1] - pos[0])
        v01s_ref = v01s.copy()
        get_distances_derivatives(v01s)
        np.testing.assert_array_equal(v01s, v01s_ref)

    @staticmethod
    def test_angles_derivatives():
        """Test `get_angles_derivatives`."""
        pos = _get_positions()
        v10s = np.atleast_2d(pos[0] - pos[1])
        v12s = np.atleast_2d(pos[2] - pos[1])
        v10s_ref = v10s.copy()
        v12s_ref = v12s.copy()
        get_angles_derivatives(v10s, v12s)
        np.testing.assert_array_equal(v10s, v10s_ref)
        np.testing.assert_array_equal(v12s, v12s_ref)

    @staticmethod
    def test_dihedrals_derivatives():
        """Test `get_dihedrals_derivatives`."""
        pos = _get_positions()
        v01s = np.atleast_2d(pos[1] - pos[0])
        v12s = np.atleast_2d(pos[2] - pos[1])
        v23s = np.atleast_2d(pos[3] - pos[2])
        v01s_ref = v01s.copy()
        v12s_ref = v12s.copy()
        v23s_ref = v23s.copy()
        get_dihedrals_derivatives(v01s, v12s, v23s)
        np.testing.assert_array_equal(v01s, v01s_ref)
        np.testing.assert_array_equal(v12s, v12s_ref)
        np.testing.assert_array_equal(v23s, v23s_ref)



import numpy as np
import pytest

from ase.dft.bee import BEEFEnsemble, ensemble, readbee


class BEECalculator:
    """Fake DFT calculator."""
    atoms = None

    def __init__(self, name):
        self.name = name

    def get_xc_functional(self):
        return self.name

    def get_nonselfconsistent_energies(self, beef_type: str) -> np.ndarray:
        n = {'mbeef': 64,
             'beefvdw': 32,
             'mbeefvdw': 28}[beef_type]
        return np.linspace(-1, 1, n)

    def get_potential_energy(self, atoms):
        return 0.0


@pytest.mark.parametrize('xc', ['mBEEF', 'BEEF-vdW', 'mBEEF-vdW'])
def test_bee(xc, testdir):
    """Check BEEF ensemble code."""
    size = 7  # size of ensemble

    # From a calculator:
    calc = BEECalculator(xc)
    ens = BEEFEnsemble(calc)
    energies = ens.get_ensemble_energies(size)
    assert energies.shape == (size,)

    # From a file:
    ens.write(f'{xc}.bee')
    e, de, contribs, _seed, xc = readbee(f'{xc}.bee', all=True)
    assert e + de == pytest.approx(energies, abs=1e-12)
    e2000 = ensemble(e, contribs, xc)
    assert e2000.shape == (2000,)

    # From data:
    ens = BEEFEnsemble(e=e, contribs=contribs, xc=xc, verbose=False)
    energies2 = ens.get_ensemble_energies(size)
    assert energies2 == pytest.approx(energies, abs=1e-12)



import numpy as np
import pytest

from ase import Atoms
from ase.build import fcc111
from ase.calculators.emt import EMT as OrigEMT
from ase.mep import DyNEB
from ase.optimize import BFGS


@pytest.mark.optimize()
@pytest.mark.slow()
def test_dynamic_neb():
    # Global counter of force evaluations:
    force_evaluations = [0]

    class EMT(OrigEMT):
        def calculate(self, *args, **kwargs):
            force_evaluations[0] += 1
            OrigEMT.calculate(self, *args, **kwargs)

    # Build Pt(111) slab with six surface atoms and add oxygen adsorbate
    initial = fcc111('Pt', size=(3, 2, 3), orthogonal=True)
    initial.center(axis=2, vacuum=10)
    oxygen = Atoms('O')
    oxygen.translate(initial[7].position + (0., 0., 3.5))
    initial.extend(oxygen)

    # EMT potential
    initial.calc = EMT()

    # Optimize initial state
    opt = BFGS(initial)
    opt.run(fmax=0.03)

    # Move oxygen adsorbate to neighboring hollow site
    final = initial.copy()
    final[18].x += 2.8
    final[18].y += 1.8

    final.calc = EMT()

    opt = BFGS(final)
    opt.run(fmax=0.03)

    # NEB with seven interior images
    images = [initial]
    for i in range(7):
        images.append(initial.copy())
    images.append(final)

    fmax = 0.03  # Same for NEB and optimizer

    for i in range(1, len(images) - 1):
        calc = EMT()
        images[i].calc = calc

    def run_NEB():
        if method == 'dyn':
            neb = DyNEB(images, fmax=fmax, dynamic_relaxation=True)
            neb.interpolate()
        elif method == 'dyn_scale':
            neb = DyNEB(images, fmax=fmax, dynamic_relaxation=True,
                        scale_fmax=6.)
            neb.interpolate()
        else:
            # Default NEB
            neb = DyNEB(images, dynamic_relaxation=False)
            neb.interpolate()

        # Optimize and check number of calculations.
        # We use a hack with a global counter to count the force evaluations:
        force_evaluations[0] = 0
        opt = BFGS(neb)
        opt.run(fmax=fmax)
        force_calls.append(force_evaluations[0])

        # Get potential energy of transition state.
        Emax.append(np.sort([image.get_potential_energy()
                             for image in images[1:-1]])[-1])

    force_calls, Emax = [], []
    for method in ['def', 'dyn', 'dyn_scale']:
        run_NEB()

    # Check force calculation count for default and dynamic NEB implementations
    print(f'\n# Force calls with default NEB: {force_calls[0]}')
    print(f'# Force calls with dynamic NEB: {force_calls[1]}')
    print('# Force calls with dynamic and scaled NEB: {}\n'.format(
        force_calls[2]))
    assert force_calls[2] < force_calls[1] < force_calls[0]

    # Assert reaction barriers are within 1 meV of default NEB
    assert (abs(Emax[1] - Emax[0]) < 1e-3)
    assert (abs(Emax[2] - Emax[0]) < 1e-3)



"""This module defines an ASE interface to Amber16.

Usage: (Tested only with Amber16, http://ambermd.org/)

Before usage, input files (infile, topologyfile, incoordfile)

"""

import subprocess

import numpy as np
from scipy.io import netcdf_file

import ase.units as units
from ase.calculators.calculator import Calculator, FileIOCalculator
from ase.io.amber import read_amber_coordinates, write_amber_coordinates


class Amber(FileIOCalculator):
    """Class for doing Amber classical MM calculations.

    Example:

    mm.in::

        Minimization with Cartesian restraints
        &cntrl
        imin=1, maxcyc=200, (invoke minimization)
        ntpr=5, (print frequency)
        &end
    """

    implemented_properties = ['energy', 'forces']
    discard_results_on_any_change = True

    def __init__(self, restart=None,
                 ignore_bad_restart_file=FileIOCalculator._deprecated,
                 label='amber', atoms=None, command=None,
                 amber_exe='sander -O ',
                 infile='mm.in', outfile='mm.out',
                 topologyfile='mm.top', incoordfile='mm.crd',
                 outcoordfile='mm_dummy.crd', mdcoordfile=None,
                 **kwargs):
        """Construct Amber-calculator object.

        Parameters
        ==========
        label: str
            Name used for all files.  May contain a directory.
        atoms: Atoms object
            Optional Atoms object to which the calculator will be
            attached.  When restarting, atoms will get its positions and
            unit-cell updated from file.
        label: str
            Prefix to use for filenames (label.in, label.txt, ...).
        amber_exe: str
            Name of the amber executable, one can add options like -O
            and other parameters here
        infile: str
            Input filename for amber, contains instuctions about the run
        outfile: str
            Logfilename for amber
        topologyfile: str
            Name of the amber topology file
        incoordfile: str
            Name of the file containing the input coordinates of atoms
        outcoordfile: str
            Name of the file containing the output coordinates of atoms
            this file is not used in case minisation/dynamics is done by ase.
            It is only relevant
            if you run MD/optimisation many steps with amber.

        """

        self.out = 'mm.log'

        self.positions = None
        self.atoms = None

        self.set(**kwargs)

        self.amber_exe = amber_exe
        self.infile = infile
        self.outfile = outfile
        self.topologyfile = topologyfile
        self.incoordfile = incoordfile
        self.outcoordfile = outcoordfile
        self.mdcoordfile = mdcoordfile

        FileIOCalculator.__init__(self, restart, ignore_bad_restart_file,
                                  label, atoms, command=command,
                                  **kwargs)

    @property
    def _legacy_default_command(self):
        command = (self.amber_exe +
                   ' -i ' + self.infile +
                   ' -o ' + self.outfile +
                   ' -p ' + self.topologyfile +
                   ' -c ' + self.incoordfile +
                   ' -r ' + self.outcoordfile)
        if self.mdcoordfile is not None:
            command = command + ' -x ' + self.mdcoordfile
        return command

    def write_input(self, atoms=None, properties=None, system_changes=None):
        """Write updated coordinates to a file."""

        FileIOCalculator.write_input(self, atoms, properties, system_changes)
        self.write_coordinates(atoms, self.incoordfile)

    def read_results(self):
        """ read energy and forces """
        self.read_energy()
        self.read_forces()

    def write_coordinates(self, atoms, filename):
        """ write amber coordinates in netCDF format,
            only rectangular unit cells are allowed"""
        write_amber_coordinates(atoms, filename)

    def read_coordinates(self, atoms):
        """Import AMBER16 netCDF restart files.

        Reads atom positions and
        velocities (if available),
        and unit cell (if available)

        This may be useful if you have run amber many steps and
        want to read new positions and velocities
        """
        # For historical reasons we edit the input atoms rather than
        # returning new atoms.
        _atoms = read_amber_coordinates(self.outcoordfile)
        atoms.cell[:] = _atoms.cell
        atoms.pbc[:] = _atoms.pbc
        atoms.positions[:] = _atoms.positions
        atoms.set_momenta(_atoms.get_momenta())

    def read_energy(self, filename='mden'):
        """ read total energy from amber file """
        with open(filename, 'r') as fd:
            lines = fd.readlines()
            blocks = []
            while 'L0' in lines[0].split()[0]:
                blocks.append(lines[0:10])
                lines = lines[10:]
                if lines == []:
                    break
        self.results['energy'] = \
            float(blocks[-1][6].split()[2]) * units.kcal / units.mol

    def read_forces(self, filename='mdfrc'):
        """ read forces from amber file """
        fd = netcdf_file(filename, 'r', mmap=False)
        try:
            forces = fd.variables['forces']
            self.results['forces'] = forces[-1, :, :] \
                / units.Ang * units.kcal / units.mol
        finally:
            fd.close()

    def set_charges(self, selection, charges, parmed_filename=None):
        """ Modify amber topology charges to contain the updated
            QM charges, needed in QM/MM.
            Using amber's parmed program to change charges.
        """
        qm_list = list(selection)
        with open(parmed_filename, 'w') as fout:
            fout.write('# update the following QM charges \n')
            for i, charge in zip(qm_list, charges):
                fout.write('change charge @' + str(i + 1) + ' ' +
                           str(charge) + ' \n')
            fout.write('# Output the topology file \n')
            fout.write('outparm ' + self.topologyfile + ' \n')
        parmed_command = ('parmed -O -i ' + parmed_filename +
                          ' -p ' + self.topologyfile +
                          ' > ' + self.topologyfile + '.log 2>&1')
        subprocess.check_call(parmed_command, shell=True, cwd=self.directory)

    def get_virtual_charges(self, atoms):
        with open(self.topologyfile, 'r') as fd:
            topology = fd.readlines()
        for n, line in enumerate(topology):
            if '%FLAG CHARGE' in line:
                chargestart = n + 2
        lines1 = topology[chargestart:(chargestart
                                       + (len(atoms) - 1) // 5 + 1)]
        mm_charges = []
        for line in lines1:
            for el in line.split():
                mm_charges.append(float(el) / 18.2223)
        charges = np.array(mm_charges)
        return charges

    def add_virtual_sites(self, positions):
        return positions  # no virtual sites

    def redistribute_forces(self, forces):
        return forces


def map(atoms, top):
    p = np.zeros((2, len(atoms)), dtype="int")

    elements = atoms.get_chemical_symbols()
    unique_elements = np.unique(atoms.get_chemical_symbols())

    for i in range(len(unique_elements)):
        idx = 0
        for j in range(len(atoms)):
            if elements[j] == unique_elements[i]:
                idx += 1
                symbol = unique_elements[i] + np.str(idx)
                for k in range(len(atoms)):
                    if top.atoms[k].name == symbol:
                        p[0, k] = j
                        p[1, j] = k
                        break
    return p


try:
    import sander
    have_sander = True
except ImportError:
    have_sander = False


class SANDER(Calculator):
    """
    Interface to SANDER using Python interface

    Requires sander Python bindings from http://ambermd.org/
    """
    implemented_properties = ['energy', 'forces']

    def __init__(self, atoms=None, label=None, top=None, crd=None,
                 mm_options=None, qm_options=None, permutation=None, **kwargs):
        if not have_sander:
            raise RuntimeError("sander Python module could not be imported!")
        Calculator.__init__(self, label, atoms)
        self.permutation = permutation
        if qm_options is not None:
            sander.setup(top, crd.coordinates, crd.box, mm_options, qm_options)
        else:
            sander.setup(top, crd.coordinates, crd.box, mm_options)

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)
        if system_changes:
            if 'energy' in self.results:
                del self.results['energy']
            if 'forces' in self.results:
                del self.results['forces']
        if 'energy' not in self.results:
            if self.permutation is None:
                crd = np.reshape(atoms.get_positions(), (1, len(atoms), 3))
            else:
                crd = np.reshape(atoms.get_positions()
                                 [self.permutation[0, :]], (1, len(atoms), 3))
            sander.set_positions(crd)
            e, f = sander.energy_forces()
            self.results['energy'] = e.tot * units.kcal / units.mol
            if self.permutation is None:
                self.results['forces'] = (np.reshape(np.array(f),
                                                     (len(atoms), 3)) *
                                          units.kcal / units.mol)
            else:
                ff = np.reshape(np.array(f), (len(atoms), 3)) * \
                    units.kcal / units.mol
                self.results['forces'] = ff[self.permutation[1, :]]
        if 'forces' not in self.results:
            if self.permutation is None:
                crd = np.reshape(atoms.get_positions(), (1, len(atoms), 3))
            else:
                crd = np.reshape(atoms.get_positions()[self.permutation[0, :]],
                                 (1, len(atoms), 3))
            sander.set_positions(crd)
            e, f = sander.energy_forces()
            self.results['energy'] = e.tot * units.kcal / units.mol
            if self.permutation is None:
                self.results['forces'] = (np.reshape(np.array(f),
                                                     (len(atoms), 3)) *
                                          units.kcal / units.mol)
            else:
                ff = np.reshape(np.array(f), (len(atoms), 3)) * \
                    units.kcal / units.mol
                self.results['forces'] = ff[self.permutation[1, :]]



"Module for displaying information about the system."


import warnings

import numpy as np

from ase.gui.i18n import _

ucellformat = """\
  {:8.3f}  {:8.3f}  {:8.3f}
  {:8.3f}  {:8.3f}  {:8.3f}
  {:8.3f}  {:8.3f}  {:8.3f}
"""


def info(gui):
    images = gui.images
    nimg = len(images)
    atoms = gui.atoms

    tokens = []

    def add(token=''):
        tokens.append(token)

    if len(atoms) < 1:
        add(_('This frame has no atoms.'))
    else:
        img = gui.frame

        if nimg == 1:
            add(_('Single image loaded.'))
        else:
            add(_('Image {} loaded (0–{}).').format(img, nimg - 1))
        add()
        add(_('Number of atoms: {}').format(len(atoms)))

        # We need to write Å³ further down, so we have no choice but to
        # use proper subscripts in the chemical formula:
        formula = atoms.get_chemical_formula()
        subscripts = dict(zip('0123456789', '₀₁₂₃₄₅₆₇₈₉'))
        pretty_formula = ''.join(subscripts.get(c, c) for c in formula)
        add(pretty_formula)

        add()
        add(_('Unit cell [Å]:'))
        add(ucellformat.format(*atoms.cell.ravel()))
        periodic = [[_('no'), _('yes')][int(periodic)]
                    for periodic in atoms.pbc]
        # TRANSLATORS: This has the form Periodic: no, no, yes
        add(_('Periodic: {}, {}, {}').format(*periodic))
        add()

        cellpar = atoms.cell.cellpar()
        add()
        add(_('Lengths [Å]: {:.3f}, {:.3f}, {:.3f}').format(*cellpar[:3]))
        add(_('Angles: {:.1f}°, {:.1f}°, {:.1f}°').format(*cellpar[3:]))

        if atoms.cell.rank == 3:
            add(_('Volume: {:.3f} Å³').format(atoms.cell.volume))

        add()

        if nimg > 1:
            if all((atoms.cell == img.cell).all() for img in images):
                add(_('Unit cell is fixed.'))
            else:
                add(_('Unit cell varies.'))

        if atoms.pbc[:2].all() and atoms.cell.rank >= 1:
            try:
                lat = atoms.cell.get_bravais_lattice()
            except RuntimeError:
                add(_('Could not recognize the lattice type'))
            except Exception:
                add(_('Unexpected error determining lattice type'))
            else:
                add(_('Reduced Bravais lattice:\n{}').format(lat))

        # Print electronic structure information if we have a calculator
        if atoms.calc:
            calc = atoms.calc

            def getresult(name, get_quantity):
                # ase/io/trajectory.py line 170 does this by using
                # the get_property(prop, atoms, allow_calculation=False)
                # so that is an alternative option.
                try:
                    if calc.calculation_required(atoms, [name]):
                        quantity = None
                    else:
                        quantity = get_quantity()
                except Exception as err:
                    quantity = None
                    errmsg = ('An error occurred while retrieving {} '
                              'from the calculator: {}'.format(name, err))
                    warnings.warn(errmsg)
                return quantity

            # SinglePointCalculators are named after the code which
            # produced the result, so this will typically list the
            # name of a code even if they are just cached results.
            add()
            from ase.calculators.singlepoint import SinglePointCalculator
            if isinstance(calc, SinglePointCalculator):
                add(_('Calculator: {} (cached)').format(calc.name))
            else:
                add(_('Calculator: {} (attached)').format(calc.name))

            energy = getresult('energy', atoms.get_potential_energy)
            forces = getresult('forces', atoms.get_forces)
            magmom = getresult('magmom', atoms.get_magnetic_moment)

            if energy is not None:
                energy_str = _('Energy: {:.3f} eV').format(energy)
                add(energy_str)

            if forces is not None:
                maxf = np.linalg.norm(forces, axis=1).max()
                forces_str = _('Max force: {:.3f} eV/Å').format(maxf)
                add(forces_str)

            if magmom is not None:
                mag_str = _('Magmom: {:.3f} µ').format(magmom)
                add(mag_str)

    return '\n'.join(tokens)



import pytest

from ase.db import connect

cmd = """
ase -T build H | ase -T run emt -o testase.json &&
ase -T build H2O | ase -T run emt -o testase.json &&
ase -T build O2 | ase -T run emt -o testase.json &&
ase -T build H2 | ase -T run emt -f 0.02 -o testase.json &&
ase -T build O2 | ase -T run emt -f 0.02 -o testase.json &&
ase -T build -x fcc Cu | ase -T run emt -E 5,1 -o testase.json &&
ase -T db -v testase.json natoms=1,Cu=1 --delete --yes &&
ase -T db -v testase.json "H>0" -k hydro=1,abc=42,foo=bar &&
ase -T db -v testase.json "H>0" --delete-keys foo"""

dbtypes = ['json', 'db', 'postgresql', 'mysql', 'mariadb']


@pytest.mark.slow()
@pytest.mark.parametrize('dbtype', dbtypes)
def test_db(dbtype, cli, testdir, get_db_name):
    def count(n, *args, **kwargs):
        m = len(list(con.select(columns=['id'], *args, **kwargs)))
        assert m == n, (m, n)

    name = get_db_name(dbtype)

    cli.shell(cmd.replace('testase.json', name))

    with connect(name) as con:
        assert con.get_atoms(H=1)[0].magmom == 1
        count(5)
        count(3, 'hydro')
        count(0, 'foo')
        count(3, abc=42)
        count(3, 'abc')
        count(0, 'abc,foo')
        count(3, 'abc,hydro')
        count(0, foo='bar')
        count(1, formula='H2')
        count(1, formula='H2O')
        count(3, 'fmax<0.1')
        count(1, '0.5<mass<1.5')
        count(5, 'energy')
        id = con.reserve(abc=7)
        assert con[id].abc == 7

        for key in ['calculator', 'energy', 'abc', 'name', 'fmax']:
            count(6, sort=key)
            count(6, sort='-' + key)

        con.delete([id])
    cli.shell(f'ase -T gui --terminal -n 3 {name}')



from ase import Atoms
from ase.io import Trajectory


def test_atoms_info_in_traj():
    info = dict(creation_date='2011-06-27',
                chemical_name='Hydrogen',
                # custom classes also works provided that it is
                # imported and pickleable...
                foo={'seven': 7})

    molecule = Atoms('H2', positions=[(0., 0., 0.), (0., 0., 1.1)], info=info)
    assert molecule.info == info

    atoms = molecule.copy()
    assert atoms.info == info

    with Trajectory('info.traj', 'w', atoms=molecule) as traj:
        traj.write()

    with Trajectory('info.traj') as traj:
        atoms = traj[-1]

    print(atoms.info)
    assert atoms.info == info



import math

from ase.utils.parsemath import eval_expression


def test_parsemath():

    param_dct = {
        "param_1": 1,
        "param_2": 4.0,
        "param_3": 478245.7586,
        "param_4": 1.58e-5,
        "param_5": -2.48757,
        "param_6": -2.0,
        "param_7": math.pi / 4.0,
    }

    expressions = [
        "3.0*param_1",
        "param_2**-2.0",
        "(param_6)**2.0",
        "param_1 / param_5",
        "param_1 + param_2 * 30.0 - param_5",
        "(param_1 + 1) / param_2",
        "sqrt(param_2)",
        "fmod(param_4, param_1)",
        "sin(param_7)",
        "cos(param_7)",
        "tan(param_7)",
        "asin(param_7)",
        "acos(param_7)",
        "atan(param_7)",
        "atan2(param_7, 2.0)",
        "hypot(3.0, 4.0)",
        "sinh(param_7)",
        "cosh(param_7)",
        "tanh(param_7)",
        "asinh(param_7)",
        "acosh(param_2)",
        "atanh(param_7)",
        "degrees(radians(param_7))",
        "log(param_3)",
        "log10(param_3)",
        "log2(param_3)",
        "abs(param_5)",
        "ceil(param_5)",
        "floor(param_5)",
        "round(param_5)",
        "exp(param_1)",
        "2.0*e",
        "pi+1",
        "tau / pi",
        "param_5 % param_6",
        "param_3 // param_6",
    ]

    solutions = [
        3 * param_dct["param_1"],
        param_dct["param_2"]**-2.0,
        (param_dct["param_6"])**2.0,
        param_dct["param_1"] / param_dct["param_5"],
        param_dct["param_1"] + param_dct["param_2"] *
        30.0 - param_dct["param_5"],
        (param_dct["param_1"] + 1) / param_dct["param_2"],
        math.sqrt(param_dct["param_2"]),
        math.fmod(param_dct["param_4"], param_dct["param_1"]),
        math.sin(param_dct["param_7"]),
        math.cos(param_dct["param_7"]),
        math.tan(param_dct["param_7"]),
        math.asin(param_dct["param_7"]),
        math.acos(param_dct["param_7"]),
        math.atan(param_dct["param_7"]),
        math.atan2(param_dct["param_7"], 2.0),
        math.hypot(3.0, 4.0),
        math.sinh(param_dct["param_7"]),
        math.cosh(param_dct["param_7"]),
        math.tanh(param_dct["param_7"]),
        math.asinh(param_dct["param_7"]),
        math.acosh(param_dct["param_2"]),
        math.atanh(param_dct["param_7"]),
        math.degrees(math.radians(param_dct["param_7"])),
        math.log(param_dct["param_3"]),
        math.log10(param_dct["param_3"]),
        math.log2(param_dct["param_3"]),
        math.fabs(param_dct["param_5"]),
        math.ceil(param_dct["param_5"]),
        math.floor(param_dct["param_5"]),
        round(param_dct["param_5"]),
        math.exp(param_dct["param_1"]),
        2.0 * math.e,
        math.pi + 1,
        2.0,
        param_dct["param_5"] % param_dct["param_6"],
    ]

    for expr, soln in zip(expressions, solutions):
        assert abs(eval_expression(expr, param_dct) - soln) < 1e-13

    try:
        eval_expression("99**99**99*99**99**99")
        raise RuntimeError(
            "This should not be reached, the parser is now vulnerable "
            "to computational time based DOS attack")
    except ValueError:
        pass

    try:
        eval_expression("e" * 10000000, {})
        raise RuntimeError(
            "This should not be reached, the parser is now vulnerable "
            "to memory based DOS attack")
    except ValueError:
        pass

    try:
        eval_expression("__import__('os').system('echo $HOME')")
        raise RuntimeError(
            "This should not be reached, "
            "the parser can execute malicious code")
    except TypeError:
        pass



"""Check if spin-symmetry is conserved"""
import pytest

from ase import Atoms


@pytest.mark.calculator('nwchem')
def test_main(factory):
    """Check is independence of alignment is conserved"""
    energies = []
    cr_atom = Atoms('Cr', positions=[(0, 0, 0)], pbc=False)
    for orientation in range(2):
        imm = 6 * (-1) ** orientation
        cr_atom.set_initial_magnetic_moments([imm])
        calculator = factory.calc(
            task='energy',
            dft=dict(convergence=dict(energy=1e-3,
                                      density=1e-2,
                                      gradient=5e-2)),
            basis='m6-31g*',
            basispar='"ao basis" spherical',
            charge=0
        )
        cr_atom.calc = calculator
        energies.append(cr_atom.get_potential_energy())
    assert abs(energies[0] - energies[1]) < 1e-9



raise ImportError(
    'The ase.ga.bulk_crossovers module has been deprecated. '
    'The same functionality is now provided by the '
    'ase.ga.cutandsplicepairing module. Please consult its documentation '
    'to verify how to e.g. initialize a CutAndSplicePairing object.')



import os
import re
import runpy
import traceback
import warnings
from os.path import join
from pathlib import Path
from stat import ST_MTIME
from subprocess import DEVNULL, CalledProcessError, check_call

from docutils import nodes
from docutils.parsers.rst.roles import set_classes


def mol_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    n = []
    t = ''
    while text:
        if text[0] == '_':
            n.append(nodes.inline(text=t))
            t = ''
            m = re.match(r'\d+', text[1:])
            if m is None:
                raise RuntimeError('Expected one or more digits after "_"')
            digits = m.group()
            n.append(nodes.subscript(text=digits))
            text = text[1 + len(digits):]
        else:
            t += text[0]
            text = text[1:]
    n.append(nodes.inline(text=t))
    return n, []


def git_role_tmpl(urlroot,
                  role,
                  rawtext, text, lineno, inliner, options={}, content=[]):
    if text[-1] == '>':
        i = text.index('<')
        name = text[:i - 1]
        text = text[i + 1:-1]
    else:
        name = text
        if name[0] == '~':
            name = name.split('/')[-1]
            text = text[1:]
        if '?' in name:
            name = name[:name.index('?')]
    # Check if the link is broken
    is_tag = text.startswith('..')  # Tags are like :git:`3.19.1 <../3.19.1>`
    path = os.path.join('..', text)
    do_exists = os.path.exists(path)
    if not (is_tag or do_exists):
        msg = f'Broken link: {rawtext}: Non-existing path: {path}'
        msg = inliner.reporter.error(msg, line=lineno)
        prb = inliner.problematic(rawtext, rawtext, msg)
        return [prb], [msg]
    ref = urlroot + text
    set_classes(options)
    node = nodes.reference(rawtext, name, refuri=ref,
                           **options)
    return [node], []


def git_role(role, rawtext, text, lineno, inliner, options={}, content=[]):
    return git_role_tmpl('https://gitlab.com/ase/ase/blob/master/',
                         role,
                         rawtext, text, lineno, inliner, options, content)


def setup(app):
    app.add_role('mol', mol_role)
    app.add_role('git', git_role)
    create_png_files()


def creates():
    """Generator for Python scripts and their output filenames."""
    for dirpath, dirnames, filenames in sorted(os.walk('.')):
        if dirpath.startswith('./build'):
            # Skip files in the build/ folder
            continue

        for filename in sorted(filenames):
            if filename.endswith('.py'):
                path = join(dirpath, filename)
                with open(path) as fd:
                    lines = fd.readlines()
                if len(lines) == 0:
                    continue
                if 'coding: utf-8' in lines[0]:
                    lines.pop(0)
                outnames = []
                for line in lines:
                    if line.startswith('# creates:'):
                        outnames.extend([file.rstrip(',')
                                         for file in line.split()[2:]])
                    else:
                        break
                if outnames:
                    yield dirpath, filename, outnames


def create_png_files(raise_exceptions=False):
    from ase.utils import workdir
    try:
        check_call(['povray', '-h'], stderr=DEVNULL)
    except (FileNotFoundError, CalledProcessError):
        warnings.warn('No POVRAY!')
        # Replace write_pov with write_png:
        from ase.io import pov
        from ase.io.png import write_png

        def write_pov(filename, atoms,
                      povray_settings={}, isosurface_data=None,
                      **generic_projection_settings):

            write_png(Path(filename).with_suffix('.png'), atoms,
                      **generic_projection_settings)

            class DummyRenderer:
                def render(self):
                    pass

            return DummyRenderer()

        pov.write_pov = write_pov

    for dir, pyname, outnames in creates():
        path = join(dir, pyname)
        t0 = os.stat(path)[ST_MTIME]
        run = False
        for outname in outnames:
            try:
                t = os.stat(join(dir, outname))[ST_MTIME]
            except OSError:
                run = True
                break
            else:
                if t < t0:
                    run = True
                    break
        if run:
            print('running:', path)
            with workdir(dir):
                import matplotlib.pyplot as plt
                plt.figure()
                try:
                    runpy.run_path(pyname)
                except KeyboardInterrupt:
                    return
                except Exception:
                    if raise_exceptions:
                        raise
                    else:
                        traceback.print_exc()

            for n in plt.get_fignums():
                plt.close(n)

            for outname in outnames:
                print(dir, outname)


def clean():
    """Remove all generated files."""
    for dir, pyname, outnames in creates():
        for outname in outnames:
            if os.path.isfile(os.path.join(dir, outname)):
                os.remove(os.path.join(dir, outname))


def visual_inspection():
    """Manually inspect generated files."""
    import subprocess
    images = []
    text = []
    pdf = []
    for dir, pyname, outnames in creates():
        for outname in outnames:
            path = os.path.join(dir, outname)
            ext = path.rsplit('.', 1)[1]
            if ext == 'pdf':
                pdf.append(path)
            elif ext in ['csv', 'txt', 'out', 'css', 'LDA', 'rst']:
                text.append(path)
            else:
                images.append(path)
    subprocess.call(['eog'] + images)
    subprocess.call(['evince'] + pdf)
    subprocess.call(['more'] + text)


if __name__ == '__main__':
    import argparse
    parser = argparse.ArgumentParser(description='Process generated files.')
    parser.add_argument('command', nargs='?', default='list',
                        choices=['list', 'inspect', 'clean', 'run'])
    args = parser.parse_args()
    if args.command == 'clean':
        clean()
    elif args.command == 'list':
        for dir, pyname, outnames in creates():
            for outname in outnames:
                print(os.path.join(dir, outname))
    elif args.command == 'run':
        create_png_files(raise_exceptions=True)
    else:
        visual_inspection()



"""
Provides LoggingCalculator class to wrap a Calculator and record
number of enery and force calls
"""

import json
import logging
import time
from typing import Any, Dict

import numpy as np

from ase.calculators.calculator import Calculator, all_properties

logger = logging.getLogger(__name__)


class LoggingCalculator(Calculator):
    """Calculator wrapper to record and plot history of energy and function
    evaluations
    """
    implemented_properties = all_properties
    default_parameters: Dict[str, Any] = {}
    name = 'LoggingCalculator'

    property_to_method_name = {
        'energy': 'get_potential_energy',
        'energies': 'get_potential_energies',
        'forces': 'get_forces',
        'stress': 'get_stress',
        'stresses': 'get_stresses'}

    def __init__(self, calculator, jsonfile=None, dumpjson=False):
        Calculator.__init__(self)
        self.calculator = calculator
        self.fmax = {}
        self.walltime = {}
        self.energy_evals = {}
        self.energy_count = {}
        self.set_label('(none)')
        if jsonfile is not None:
            self.read_json(jsonfile)
        self.dumpjson = dumpjson

    def calculate(self, atoms, properties, system_changes):
        Calculator.calculate(self, atoms, properties, system_changes)

        if isinstance(self.calculator, Calculator):
            results = [self.calculator.get_property(prop, atoms)
                       for prop in properties]
        else:
            results = []
            for prop in properties:
                method_name = self.property_to_method_name[prop]
                method = getattr(self.calculator, method_name)
                results.append(method(atoms))

        if 'energy' in properties or 'energies' in properties:
            self.energy_evals.setdefault(self.label, 0)
            self.energy_evals[self.label] += 1
            try:
                energy = results[properties.index('energy')]
            except IndexError:
                energy = sum(results[properties.index('energies')])
            logger.info('energy call count=%d energy=%.3f',
                        self.energy_evals[self.label], energy)
        self.results = dict(zip(properties, results))

        if 'forces' in self.results:
            fmax = self.fmax.setdefault(self.label, [])
            walltime = self.walltime.setdefault(self.label, [])
            forces = self.results['forces'].copy()
            energy_count = self.energy_count.setdefault(self.label, [])
            energy_evals = self.energy_evals.setdefault(self.label, 0)
            energy_count.append(energy_evals)
            for constraint in atoms.constraints:
                constraint.adjust_forces(atoms, forces)
            fmax.append(abs(forces).max())
            walltime.append(time.time())
            logger.info('force call fmax=%.3f', fmax[-1])

        if self.dumpjson:
            self.write_json('dump.json')

    def write_json(self, filename):
        with open(filename, 'w') as fd:
            json.dump({'fmax': self.fmax,
                       'walltime': self.walltime,
                       'energy_evals': self.energy_evals,
                       'energy_count': self.energy_count}, fd)

    def read_json(self, filename, append=False, label=None):
        with open(filename) as fd:
            dct = json.load(fd)

        labels = dct['fmax'].keys()
        if label is not None and len(labels) == 1:
            for key in ('fmax', 'walltime', 'energy_evals', 'energy_count'):
                dct[key][label] = dct[key][labels[0]]
                del dct[key][labels[0]]
        if not append:
            self.fmax = {}
            self.walltime = {}
            self.energy_evals = {}
            self.energy_count = {}
        self.fmax.update(dct['fmax'])
        self.walltime.update(dct['walltime'])
        self.energy_evals.update(dct['energy_evals'])
        self.energy_count.update(dct['energy_count'])

    def tabulate(self):
        fmt1 = '%-10s %10s %10s %8s'
        title = fmt1 % ('Label', '# Force', '# Energy', 'Walltime/s')
        print(title)
        print('-' * len(title))
        fmt2 = '%-10s %10d %10d %8.2f'
        for label in sorted(self.fmax.keys()):
            print(fmt2 % (label, len(self.fmax[label]),
                          len(self.energy_count[label]),
                          self.walltime[label][-1] - self.walltime[label][0]))

    def plot(self, fmaxlim=(1e-2, 1e2), forces=True, energy=True,
             walltime=True,
             markers=None, labels=None, **kwargs):
        import matplotlib.pyplot as plt

        if markers is None:
            markers = [c + s for c in ['r', 'g', 'b', 'c', 'm', 'y', 'k']
                       for s in ['.-', '.--']]
        nsub = sum([forces, energy, walltime])
        nplot = 0

        if labels is not None:
            fmax_values = [v for (k, v) in sorted(zip(self.fmax.keys(),
                                                      self.fmax.values()))]
            self.fmax = dict(zip(labels, fmax_values))

            energy_count_values = [v for (k, v) in
                                   sorted(zip(self.energy_count.keys(),
                                              self.energy_count.values()))]
            self.energy_count = dict(zip(labels, energy_count_values))

            walltime_values = [v for (k, v) in
                               sorted(zip(self.walltime.keys(),
                                          self.walltime.values()))]
            self.walltime = dict(zip(labels, walltime_values))

        if forces:
            nplot += 1
            plt.subplot(nsub, 1, nplot)
            for label, color in zip(sorted(self.fmax.keys()), markers):
                fmax = np.array(self.fmax[label])
                idx = np.arange(len(fmax))
                plt.semilogy(idx, fmax, color, label=label, **kwargs)

            plt.xlabel('Number of force evaluations')
            plt.ylabel('Maximum force / eV/A')
            plt.ylim(*fmaxlim)
            plt.legend()

        if energy:
            nplot += 1
            plt.subplot(nsub, 1, nplot)
            for label, color in zip(sorted(self.energy_count.keys()), markers):
                energy_count = np.array(self.energy_count[label])
                fmax = np.array(self.fmax[label])
                plt.semilogy(energy_count, fmax, color, label=label, **kwargs)

            plt.xlabel('Number of energy evaluations')
            plt.ylabel('Maximum force / eV/A')
            plt.ylim(*fmaxlim)
            plt.legend()

        if walltime:
            nplot += 1
            plt.subplot(nsub, 1, nplot)
            for label, color in zip(sorted(self.walltime.keys()), markers):
                walltime = np.array(self.walltime[label])
                fmax = np.array(self.fmax[label])
                walltime -= walltime[0]
                plt.semilogy(walltime, fmax, color, label=label, **kwargs)

            plt.xlabel('Walltime / s')
            plt.ylabel('Maximum force / eV/A')
            plt.ylim(*fmaxlim)
            plt.legend()

        plt.subplots_adjust(hspace=0.33)



import numpy as np

delta = 1e-10


def wulff_construction(symbol, surfaces, energies, size, structure,
                       rounding='closest', latticeconstant=None,
                       debug=False, maxiter=100):
    """Create a cluster using the Wulff construction.

    A cluster is created with approximately the number of atoms
    specified, following the Wulff construction, i.e. minimizing the
    surface energy of the cluster.

    Parameters
    ----------
    symbol : str or int
        The chemical symbol (or atomic number) of the desired element.

    surfaces : list
        A list of surfaces. Each surface is an (h, k, l) tuple or list of
        integers.

    energies : list
        A list of surface energies for the surfaces.

    size : int
        The desired number of atoms.

    structure : {'fcc', bcc', 'sc'}
        The desired crystal structure.

    rounding : {'closest', 'above', 'below'}
        Specifies what should be done if no Wulff construction corresponds
        to exactly the requested number of atoms. 'above', 'below', and
        'closest' mean that the nearest cluster above or below - or the
        closest one - is created instead.

    latticeconstant : float (optional)
        The lattice constant. If not given, extracted from `ase.data`.

    debug : bool, default False
        If True, information about the iteration towards the right cluster
        size is printed.
    """

    if debug:
        print('Wulff: Aiming for cluster with %i atoms (%s)' %
              (size, rounding))

        if rounding not in ['above', 'below', 'closest']:
            raise ValueError(f'Invalid rounding: {rounding}')

    # Interpret structure, if it is a string.
    if isinstance(structure, str):
        if structure == 'fcc':
            from ase.cluster.cubic import FaceCenteredCubic as structure
        elif structure == 'bcc':
            from ase.cluster.cubic import BodyCenteredCubic as structure
        elif structure == 'sc':
            from ase.cluster.cubic import SimpleCubic as structure
        elif structure == 'hcp':
            from ase.cluster.hexagonal import HexagonalClosedPacked as structure
        elif structure == 'graphite':
            from ase.cluster.hexagonal import Graphite as structure
        else:
            error = f'Crystal structure {structure} is not supported.'
            raise NotImplementedError(error)

    # Check number of surfaces
    nsurf = len(surfaces)
    if len(energies) != nsurf:
        raise ValueError('The energies array should contain %d values.'
                         % (nsurf,))

    # Copy energies array so it is safe to modify it
    energies = np.array(energies)

    # We should check that for each direction, the surface energy plus
    # the energy in the opposite direction is positive.  But this is
    # very difficult in the general case!

    # Before starting, make a fake cluster just to extract the
    # interlayer distances in the relevant directions, and use these
    # to "renormalize" the surface energies such that they can be used
    # to convert to number of layers instead of to distances.
    atoms = structure(symbol, surfaces, 5 * np.ones(len(surfaces), int),
                      latticeconstant=latticeconstant)
    for i, s in enumerate(surfaces):
        d = atoms.get_layer_distance(s)
        energies[i] /= d

    # First guess a size that is not too large.
    wanted_size = size ** (1.0 / 3.0)
    max_e = max(energies)
    factor = wanted_size / max_e
    atoms, layers = make_atoms(symbol, surfaces, energies, factor, structure,
                               latticeconstant)
    if len(atoms) == 0:
        # Probably the cluster is very flat
        if debug:
            print('First try made an empty cluster, trying again.')
        factor = 1 / energies.min()
        atoms, layers = make_atoms(symbol, surfaces, energies, factor,
                                   structure, latticeconstant)
        if len(atoms) == 0:
            raise RuntimeError('Failed to create a finite cluster.')

    # Second guess: scale to get closer.
    old_factor = factor
    old_layers = layers
    old_atoms = atoms
    factor *= (size / len(atoms))**(1.0 / 3.0)
    atoms, layers = make_atoms(symbol, surfaces, energies, factor,
                               structure, latticeconstant)
    if len(atoms) == 0:
        print('Second guess gave an empty cluster, discarding it.')
        atoms = old_atoms
        factor = old_factor
        layers = old_layers
    else:
        del old_atoms

    # Find if the cluster is too small or too large (both means perfect!)
    below = above = None
    if len(atoms) <= size:
        below = atoms
    if len(atoms) >= size:
        above = atoms

    # Now iterate towards the right cluster
    iter = 0
    while (below is None or above is None):
        if len(atoms) < size:
            # Find a larger cluster
            if debug:
                print('Making a larger cluster.')
            factor = ((layers + 0.5 + delta) / energies).min()
            atoms, new_layers = make_atoms(symbol, surfaces, energies, factor,
                                           structure, latticeconstant)
            assert (new_layers - layers).max() == 1
            assert (new_layers - layers).min() >= 0
            layers = new_layers
        else:
            # Find a smaller cluster
            if debug:
                print('Making a smaller cluster.')
            factor = ((layers - 0.5 - delta) / energies).max()
            atoms, new_layers = make_atoms(symbol, surfaces, energies, factor,
                                           structure, latticeconstant)
            assert (new_layers - layers).max() <= 0
            assert (new_layers - layers).min() == -1
            layers = new_layers
        if len(atoms) <= size:
            below = atoms
        if len(atoms) >= size:
            above = atoms
        iter += 1
        if iter == maxiter:
            raise RuntimeError('Runaway iteration.')
    if rounding == 'below':
        if debug:
            print('Choosing smaller cluster with %i atoms' % len(below))
        return below
    elif rounding == 'above':
        if debug:
            print('Choosing larger cluster with %i atoms' % len(above))
        return above
    else:
        assert rounding == 'closest'
        if (len(above) - size) < (size - len(below)):
            atoms = above
        else:
            atoms = below
        if debug:
            print('Choosing closest cluster with %i atoms' % len(atoms))
        return atoms


def make_atoms(symbol, surfaces, energies, factor, structure, latticeconstant):
    layers1 = factor * np.array(energies)
    layers = np.round(layers1).astype(int)
    atoms = structure(symbol, surfaces, layers,
                      latticeconstant=latticeconstant)
    return (atoms, layers)



def write_findsym(fileobj, images):
    symbols = images[0].get_chemical_symbols()
    natoms = len(symbols)

    for atoms in images:
        formula = atoms.get_chemical_symbols()
        accuracy = 1.0e-4

        # Write Comment
        fileobj.write(f'{formula}\n')
        fileobj.write('%f   accuracy\n' % accuracy)
        fileobj.write('1    vectors in cartesian coordinates\n')

        # Write cartesian coordinates of vectors
        for x, y, z in atoms.cell:
            fileobj.write(f'{x:22.15f} {y:22.15f} {z:22.15f}\n')

        fileobj.write('1    no known centering\n')

        fileobj.write('1 0 0 \n')
        fileobj.write('0 1 0 \n')
        fileobj.write('0 0 1 \n')

        fileobj.write('%d\n' % natoms)

        numbers = atoms.get_atomic_numbers()
        for n in numbers:
            fileobj.write('%d ' % (n))

        fileobj.write('\n')

        for x, y, z in atoms.get_positions():
            fileobj.write(f'{x:22.15f} {y:22.15f} {z:22.15f}\n')



import pytest

from ase.lattice.cubic import FaceCenteredCubic
from ase.lattice.hexagonal import HexagonalClosedPacked


def test_miller_lindep():
    with pytest.raises(ValueError):
        # The Miller indices of the surfaces are linearly dependent
        FaceCenteredCubic(symbol='Cu',
                          miller=[[1, 1, 0], [1, 1, 0], [0, 0, 1]])


def test_fcc_ok():
    atoms = FaceCenteredCubic(symbol='Cu',
                              miller=[[1, 1, 0], [0, 1, 0], [0, 0, 1]])
    print(atoms.get_cell())


@pytest.mark.parametrize('directions', [
    [[1, 1, 0], [1, 1, 0], [0, 0, 1]],
    [[1, 1, 0], [1, 0, 0], [0, 1, 0]]
])
def test_fcc_directions_linearly_dependent(directions):
    # The directions spanning the unit cell are linearly dependent
    with pytest.raises(ValueError):
        FaceCenteredCubic(symbol='Cu', directions=directions)


def test_fcc_directions_ok():
    atoms = FaceCenteredCubic(symbol='Cu',
                              directions=[[1, 1, 0], [0, 1, 0], [0, 0, 1]])
    print(atoms.get_cell())


def test_hcp_miller_lienarly_dependent():
    with pytest.raises((ValueError, NotImplementedError)):
        # The Miller indices of the surfaces are linearly dependent
        HexagonalClosedPacked(symbol='Mg',
                              miller=[[1, -1, 0, 0],
                                      [1, 0, -1, 0],
                                      [0, 1, -1, 0]])

    # This one should be OK
    #
    # It is not!  The miller argument is broken in hexagonal crystals!
    #
    # atoms = HexagonalClosedPacked(symbol='Mg',
    #                               miller=[[1, -1, 0, 0],
    #                                       [1, 0, -1, 0],
    #                                       [0, 0, 0, 1]])
    # print(atoms.get_cell())


def test_hcp_cell_linearly_dependent():
    with pytest.raises(ValueError):
        # The directions spanning the unit cell are linearly dependent
        HexagonalClosedPacked(symbol='Mg',
                              directions=[[1, -1, 0, 0],
                                          [1, 0, -1, 0],
                                          [0, 1, -1, 0]])


def test_hcp():
    atoms = HexagonalClosedPacked(symbol='Mg',
                                  directions=[[1, -1, 0, 0],
                                              [1, 0, -1, 0],
                                              [0, 0, 0, 1]])
    print(atoms.get_cell())



import numpy as np

from ase.build import bulk
from ase.calculators.emt import EMT
from ase.calculators.fd import FiniteDifferenceCalculator


def test_fd():
    """Test `FiniteDifferenceCalculator`."""
    atoms = bulk('Cu', cubic=True)
    atoms.rattle(0.1)

    atoms.calc = EMT()
    energy_analytical = atoms.get_potential_energy()
    forces_analytical = atoms.get_forces()
    stress_analytical = atoms.get_stress()

    atoms.calc = FiniteDifferenceCalculator(EMT())
    energy_numerical = atoms.get_potential_energy()
    forces_numerical = atoms.get_forces()
    stress_numerical = atoms.get_stress()

    # check if `numerical` energy is exactly equal to `analytical`
    assert energy_numerical == energy_analytical

    # check if `numerical` forces are *not* exactly equal to `analytical`
    assert np.any(forces_numerical != forces_analytical)
    assert np.any(stress_numerical != stress_analytical)

    np.testing.assert_allclose(forces_numerical, forces_analytical)
    np.testing.assert_allclose(stress_numerical, stress_analytical)



"""Tests for legacy and modern NumPy PRNGs."""
import numpy as np
import pytest

from ase.build import bulk
from ase.calculators.emt import EMT
from ase.md import Andersen
from ase.units import fs


@pytest.fixture(name='atoms')
def _fixture_atoms():
    atoms = bulk('Cu') * (2, 2, 2)
    atoms.calc = EMT()
    return atoms


@pytest.mark.parametrize(
    'rng',
    [np.random.RandomState(0), np.random.default_rng(42)],
)
def test_andersen(atoms, rng):
    """Test legacy and modern NumPy PRNGs for `Andersen`."""
    with Andersen(
        atoms,
        timestep=1.0 / fs,
        temperature_K=1000.0,
        andersen_prob=0.01,
        rng=rng,
    ) as md:
        md.run(5)



from pathlib import Path

from ase.calculators.abc import GetOutputsMixin
from ase.calculators.calculator import FileIOCalculator
from ase.io import write
from ase.io.elk import ElkReader


class ELK(FileIOCalculator, GetOutputsMixin):
    _legacy_default_command = 'elk > elk.out'
    implemented_properties = ['energy', 'forces']
    ignored_changes = {'pbc'}
    discard_results_on_any_change = True

    fileio_rules = FileIOCalculator.ruleset(
        stdout_name='elk.out')

    def __init__(self, **kwargs):
        """Construct ELK calculator.

        The keyword arguments (kwargs) can be one of the ASE standard
        keywords: 'xc', 'kpts' and 'smearing' or any of ELK'
        native keywords.
        """

        super().__init__(**kwargs)

    def write_input(self, atoms, properties=None, system_changes=None):
        FileIOCalculator.write_input(self, atoms, properties, system_changes)

        parameters = dict(self.parameters)
        if 'forces' in properties:
            parameters['tforce'] = True

        directory = Path(self.directory)
        write(directory / 'elk.in', atoms, parameters=parameters,
              format='elk-in')

    def read_results(self):
        from ase.outputs import Properties
        reader = ElkReader(self.directory)
        dct = dict(reader.read_everything())

        converged = dct.pop('converged')
        if not converged:
            raise RuntimeError('Did not converge')

        # (Filter results thorugh Properties for error detection)
        props = Properties(dct)
        self.results = dict(props)

    def _outputmixin_get_results(self):
        return self.results



'''
Conversion of text from a Crystallographic Information File (CIF) format to
unicode. CIF text is neither unicode nor bibtex/latex code.

Rules for character formatting in CIF files are specified at:
https://www.iucr.org/resources/cif/spec/version1.1/semantics
'''

import html
import re

subs_dict = {
    '\r': '',            # Windows line ending
    '\t': ' ',           # tabs

    r'\a': '\u03b1',    # alpha
    r'\b': '\u03b2',    # beta
    r'\g': '\u03b3',    # gamma
    r'\d': '\u03b4',    # delta
    r'\e': '\u03b5',    # epsilon
    r'\z': '\u03b6',    # zeta
    r'\h': '\u03b7',    # eta
    r'\q': '\u03b8',    # theta
    r'\i': '\u03b9',    # iota
    r'\k': '\u03ba',    # kappa
    r'\l': '\u03bb',    # lambda
    r'\m': '\u03bc',    # mu
    r'\n': '\u03bd',    # nu
    r'\x': '\u03be',    # xi
    r'\o': '\u03bf',    # omicron
    r'\p': '\u03c0',    # pi
    r'\r': '\u03c1',    # rho
    r'\s': '\u03c3',    # sigma
    r'\t': '\u03c4',    # tau
    r'\u': '\u03c5',    # upsilon
    r'\f': '\u03c6',    # phi
    r'\c': '\u03c7',    # chi
    r'\y': '\u03c8',    # psi
    r'\w': '\u03c9',    # omega
    r'\A': '\u0391',    # Alpha
    r'\B': '\u0392',    # Beta
    r'\G': '\u0393',    # Gamma
    r'\D': '\u0394',    # Delta
    r'\E': '\u0395',    # Epsilon
    r'\Z': '\u0396',    # Zeta
    r'\H': '\u0397',    # Eta
    r'\Q': '\u0398',    # Theta
    r'\I': '\u0399',    # Ioto
    r'\K': '\u039a',    # Kappa
    r'\L': '\u039b',    # Lambda
    r'\M': '\u039c',    # Mu
    r'\N': '\u039d',    # Nu
    r'\X': '\u039e',    # Xi
    r'\O': '\u039f',    # Omicron
    r'\P': '\u03a0',    # Pi
    r'\R': '\u03a1',    # Rho
    r'\S': '\u03a3',    # Sigma
    r'\T': '\u03a4',    # Tau
    r'\U': '\u03a5',    # Upsilon
    r'\F': '\u03a6',    # Phi
    r'\C': '\u03a7',    # Chi
    r'\Y': '\u03a8',    # Psi
    r'\W': '\u03a9',    # Omega

    r'\%a': '\u00e5',   # a-ring
    r'\/o': '\u00f8',   # o-slash
    r'\?i': '\u0131',   # dotless i
    r'\/l': '\u0142',   # Polish l
    r'\&s': '\u00df',   # German eszett
    r'\/d': '\u0111',   # barred d

    r'\%A': '\u00c5',   # A-ring
    r'\/O': '\u00d8',   # O-slash
    r'\?I': 'I',         # dotless I
    r'\/L': '\u0141',   # Polish L
    r'\&S': '\u1e9e',   # German Eszett
    r'\/D': '\u0110',   # barred D

    r'\%': '\u00b0',           # degree
    r'--': '\u2013',           # dash
    r'---': '\u2014',          # single bond
    r'\\db': '\u003d',         # double bond
    r'\\tb': '\u2261',         # triple bond
    r'\\ddb': '\u2248',        # delocalized double bond
    r'\\sim': '~',
    r'\\simeq': '\u2243',
    r'\\infty': '\u221e',      # infinity

    r'\\times': '\u00d7',
    r'+-': '\u00b1',           # plusminus
    r'-+': '\u2213',           # minusplus
    r'\\square': '\u25a0',
    r'\\neq': '\u2660',
    r'\\rangle': '\u3009',
    r'\\langle': '\u3008',
    r'\\rightarrow': '\u2192',
    r'\\leftarrow': '\u2190',

    r"\'A": '\u00c1',  # A acute
    r"\'E": '\u00c9',  # E acute
    r"\'I": '\u00cd',  # I acute
    r"\'O": '\u00d3',  # O acute
    r"\'U": '\u00da',  # U acute
    r"\'Y": '\u00dd',  # Y acute
    r"\'a": '\u00e1',  # a acute
    r"\'e": '\u00e9',  # e acute
    r"\'i": '\u00ed',  # i acute
    r"\'o": '\u00f3',  # o acute
    r"\'u": '\u00fa',  # u acute
    r"\'y": '\u00fd',  # y acute
    r"\'C": '\u0106',  # C acute
    r"\'c": '\u0107',  # c acute
    r"\'L": '\u0139',  # L acute
    r"\'l": '\u013a',  # l acute
    r"\'N": '\u0143',  # N acute
    r"\'n": '\u0144',  # n acute
    r"\'R": '\u0154',  # R acute
    r"\'r": '\u0155',  # r acute
    r"\'S": '\u015a',  # S acute
    r"\'s": '\u015b',  # s acute
    r"\'Z": '\u0179',  # Z acute
    r"\'z": '\u017a',  # z acute
    r"\'G": '\u01f4',  # G acute
    r"\'g": '\u01f5',  # g acute
    r"\'K": '\u1e30',  # K acute
    r"\'k": '\u1e31',  # k acute
    r"\'M": '\u1e3e',  # M acute
    r"\'m": '\u1e3f',  # m acute
    r"\'P": '\u1e54',  # P acute
    r"\'p": '\u1e55',  # p acute
    r"\'W": '\u1e82',  # W acute
    r"\'w": '\u1e83',  # w acute
    r'\;A': '\u0104',  # A ogonek
    r'\;a': '\u0105',  # a ogonek
    r'\;E': '\u0118',  # E ogonek
    r'\;e': '\u0119',  # e ogonek
    r'\;I': '\u012e',  # I ogonek
    r'\;i': '\u012f',  # i ogonek
    r'\;U': '\u0172',  # U ogonek
    r'\;u': '\u0173',  # u ogonek
    r'\;O': '\u01ea',  # O ogonek
    r'\;o': '\u01eb',  # o ogonek
    r'\.C': '\u010a',  # C dot above
    r'\.c': '\u010b',  # c dot above
    r'\.E': '\u0116',  # E dot above
    r'\.e': '\u0117',  # e dot above
    r'\.G': '\u0120',  # G dot above
    r'\.g': '\u0121',  # g dot above
    r'\.I': '\u0130',  # I dot above
    r'\.Z': '\u017b',  # Z dot above
    r'\.z': '\u017c',  # z dot above
    r'\.A': '\u0226',  # A dot above
    r'\.a': '\u0227',  # a dot above
    r'\.O': '\u022e',  # O dot above
    r'\.o': '\u022f',  # o dot above
    r'\.B': '\u1e02',  # B dot above
    r'\.b': '\u1e03',  # b dot above
    r'\.D': '\u1e0a',  # D dot above
    r'\.d': '\u1e0b',  # d dot above
    r'\.F': '\u1e1e',  # F dot above
    r'\.f': '\u1e1f',  # f dot above
    r'\.H': '\u1e22',  # H dot above
    r'\.h': '\u1e23',  # h dot above
    r'\.M': '\u1e40',  # M dot above
    r'\.m': '\u1e41',  # m dot above
    r'\.N': '\u1e44',  # N dot above
    r'\.n': '\u1e45',  # n dot above
    r'\.P': '\u1e56',  # P dot above
    r'\.p': '\u1e57',  # p dot above
    r'\.R': '\u1e58',  # R dot above
    r'\.r': '\u1e59',  # r dot above
    r'\.S': '\u1e60',  # S dot above
    r'\.s': '\u1e61',  # s dot above
    r'\.T': '\u1e6a',  # T dot above
    r'\.t': '\u1e6b',  # t dot above
    r'\.W': '\u1e86',  # W dot above
    r'\.w': '\u1e87',  # w dot above
    r'\.X': '\u1e8a',  # X dot above
    r'\.x': '\u1e8b',  # x dot above
    r'\.Y': '\u1e8e',  # Y dot above
    r'\.y': '\u1e8f',  # y dot above
    r'\(A': '\u0102',  # A breve
    r'\(a': '\u0103',  # a breve
    r'\(E': '\u0114',  # E breve
    r'\(e': '\u0115',  # e breve
    r'\(G': '\u011e',  # G breve
    r'\(g': '\u011f',  # g breve
    r'\(I': '\u012c',  # I breve
    r'\(i': '\u012d',  # i breve
    r'\(O': '\u014e',  # O breve
    r'\(o': '\u014f',  # o breve
    r'\(U': '\u016c',  # U breve
    r'\(u': '\u016d',  # u breve
    r'\=A': '\u0100',  # A macron
    r'\=a': '\u0101',  # a macron
    r'\=E': '\u0112',  # E macron
    r'\=e': '\u0113',  # e macron
    r'\=I': '\u012a',  # I macron
    r'\=i': '\u012b',  # i macron
    r'\=O': '\u014c',  # O macron
    r'\=o': '\u014d',  # o macron
    r'\=U': '\u016a',  # U macron
    r'\=u': '\u016b',  # u macron
    r'\=Y': '\u0232',  # Y macron
    r'\=y': '\u0233',  # y macron
    r'\=G': '\u1e20',  # G macron
    r'\=g': '\u1e21',  # g macron
    r'\^A': '\u00c2',  # A circumflex
    r'\^E': '\u00ca',  # E circumflex
    r'\^I': '\u00ce',  # I circumflex
    r'\^O': '\u00d4',  # O circumflex
    r'\^U': '\u00db',  # U circumflex
    r'\^a': '\u00e2',  # a circumflex
    r'\^e': '\u00ea',  # e circumflex
    r'\^i': '\u00ee',  # i circumflex
    r'\^o': '\u00f4',  # o circumflex
    r'\^u': '\u00fb',  # u circumflex
    r'\^C': '\u0108',  # C circumflex
    r'\^c': '\u0109',  # c circumflex
    r'\^G': '\u011c',  # G circumflex
    r'\^g': '\u011d',  # g circumflex
    r'\^H': '\u0124',  # H circumflex
    r'\^h': '\u0125',  # h circumflex
    r'\^J': '\u0134',  # J circumflex
    r'\^j': '\u0135',  # j circumflex
    r'\^S': '\u015c',  # S circumflex
    r'\^s': '\u015d',  # s circumflex
    r'\^W': '\u0174',  # W circumflex
    r'\^w': '\u0175',  # w circumflex
    r'\^Y': '\u0176',  # Y circumflex
    r'\^y': '\u0177',  # y circumflex
    r'\^Z': '\u1e90',  # Z circumflex
    r'\^z': '\u1e91',  # z circumflex
    r'\"A': '\u00c4',  # A diaeresis
    r'\"E': '\u00cb',  # E diaeresis
    r'\"I': '\u00cf',  # I diaeresis
    r'\"O': '\u00d6',  # O diaeresis
    r'\"U': '\u00dc',  # U diaeresis
    r'\"a': '\u00e4',  # a diaeresis
    r'\"e': '\u00eb',  # e diaeresis
    r'\"i': '\u00ef',  # i diaeresis
    r'\"o': '\u00f6',  # o diaeresis
    r'\"u': '\u00fc',  # u diaeresis
    r'\"y': '\u00ff',  # y diaeresis
    r'\"Y': '\u0178',  # Y diaeresis
    r'\"H': '\u1e26',  # H diaeresis
    r'\"h': '\u1e27',  # h diaeresis
    r'\"W': '\u1e84',  # W diaeresis
    r'\"w': '\u1e85',  # w diaeresis
    r'\"X': '\u1e8c',  # X diaeresis
    r'\"x': '\u1e8d',  # x diaeresis
    r'\"t': '\u1e97',  # t diaeresis
    r'\~A': '\u00c3',  # A tilde
    r'\~N': '\u00d1',  # N tilde
    r'\~O': '\u00d5',  # O tilde
    r'\~a': '\u00e3',  # a tilde
    r'\~n': '\u00f1',  # n tilde
    r'\~o': '\u00f5',  # o tilde
    r'\~I': '\u0128',  # I tilde
    r'\~i': '\u0129',  # i tilde
    r'\~U': '\u0168',  # U tilde
    r'\~u': '\u0169',  # u tilde
    r'\~V': '\u1e7c',  # V tilde
    r'\~v': '\u1e7d',  # v tilde
    r'\~E': '\u1ebc',  # E tilde
    r'\~e': '\u1ebd',  # e tilde
    r'\~Y': '\u1ef8',  # Y tilde
    r'\~y': '\u1ef9',  # y tilde
    r'\<C': '\u010c',  # C caron
    r'\<c': '\u010d',  # c caron
    r'\<D': '\u010e',  # D caron
    r'\<d': '\u010f',  # d caron
    r'\<E': '\u011a',  # E caron
    r'\<e': '\u011b',  # e caron
    r'\<L': '\u013d',  # L caron
    r'\<l': '\u013e',  # l caron
    r'\<N': '\u0147',  # N caron
    r'\<n': '\u0148',  # n caron
    r'\<R': '\u0158',  # R caron
    r'\<r': '\u0159',  # r caron
    r'\<S': '\u0160',  # S caron
    r'\<s': '\u0161',  # s caron
    r'\<T': '\u0164',  # T caron
    r'\<t': '\u0165',  # t caron
    r'\<Z': '\u017d',  # Z caron
    r'\<z': '\u017e',  # z caron
    r'\<A': '\u01cd',  # A caron
    r'\<a': '\u01ce',  # a caron
    r'\<I': '\u01cf',  # I caron
    r'\<i': '\u01d0',  # i caron
    r'\<O': '\u01d1',  # O caron
    r'\<o': '\u01d2',  # o caron
    r'\<U': '\u01d3',  # U caron
    r'\<u': '\u01d4',  # u caron
    r'\<G': '\u01e6',  # G caron
    r'\<g': '\u01e7',  # g caron
    r'\<K': '\u01e8',  # K caron
    r'\<k': '\u01e9',  # k caron
    r'\<j': '\u01f0',  # j caron
    r'\<H': '\u021e',  # H caron
    r'\<h': '\u021f',  # h caron
    r'\>O': '\u0150',  # O double acute
    r'\>o': '\u0151',  # o double acute
    r'\>U': '\u0170',  # U double acute
    r'\>u': '\u0171',  # u double acute
    r'\,C': '\u00c7',  # C cedilla
    r'\,c': '\u00e7',  # c cedilla
    r'\,G': '\u0122',  # G cedilla
    r'\,g': '\u0123',  # g cedilla
    r'\,K': '\u0136',  # K cedilla
    r'\,k': '\u0137',  # k cedilla
    r'\,L': '\u013b',  # L cedilla
    r'\,l': '\u013c',  # l cedilla
    r'\,N': '\u0145',  # N cedilla
    r'\,n': '\u0146',  # n cedilla
    r'\,R': '\u0156',  # R cedilla
    r'\,r': '\u0157',  # r cedilla
    r'\,S': '\u015e',  # S cedilla
    r'\,s': '\u015f',  # s cedilla
    r'\,T': '\u0162',  # T cedilla
    r'\,t': '\u0163',  # t cedilla
    r'\,E': '\u0228',  # E cedilla
    r'\,e': '\u0229',  # e cedilla
    r'\,D': '\u1e10',  # D cedilla
    r'\,d': '\u1e11',  # d cedilla
    r'\,H': '\u1e28',  # H cedilla
    r'\,h': '\u1e29',  # h cedilla
    r'\`A': '\u00c0',  # A grave
    r'\`E': '\u00c8',  # E grave
    r'\`I': '\u00cc',  # I grave
    r'\`O': '\u00d2',  # O grave
    r'\`U': '\u00d9',  # U grave
    r'\`a': '\u00e0',  # a grave
    r'\`e': '\u00e8',  # e grave
    r'\`i': '\u00ec',  # i grave
    r'\`o': '\u00f2',  # o grave
    r'\`u': '\u00f9',  # u grave
    r'\`N': '\u01f8',  # N grave
    r'\`n': '\u01f9',  # n grave
    r'\`W': '\u1e80',  # W grave
    r'\`w': '\u1e81',  # w grave
    r'\`Y': '\u1ef2',  # Y grave
    r'\`y': '\u1ef3',  # y grave
}

superscript_dict = {
    '0': '\u2070',  # superscript 0
    '1': '\u00b9',  # superscript 1
    '2': '\u00b2',  # superscript 2
    '3': '\u00b3',  # superscript 3
    '4': '\u2074',  # superscript 4
    '5': '\u2075',  # superscript 5
    '6': '\u2076',  # superscript 6
    '7': '\u2077',  # superscript 7
    '8': '\u2078',  # superscript 8
    '9': '\u2079',  # superscript 9
}

subscript_dict = {
    '0': '\u2080',  # subscript 0
    '1': '\u2081',  # subscript 1
    '2': '\u2082',  # subscript 2
    '3': '\u2083',  # subscript 3
    '4': '\u2084',  # subscript 4
    '5': '\u2085',  # subscript 5
    '6': '\u2086',  # subscript 6
    '7': '\u2087',  # subscript 7
    '8': '\u2088',  # subscript 8
    '9': '\u2089',  # subscript 9
}


def replace_subscript(s: str, subscript=True) -> str:

    target = '~'
    rdict = subscript_dict
    if not subscript:
        target = '^'
        rdict = superscript_dict

    replaced = []
    inside = False
    for char in s:
        if char == target:
            inside = not inside
        elif not inside:
            replaced += [char]
        # note: do not use char.isdigit - this also matches (sub/super)scripts
        elif char in rdict:
            replaced += [rdict[char]]
        else:
            replaced += [char]

    return ''.join(replaced)


def multiple_replace(text: str, adict) -> str:
    rx = re.compile('|'.join(map(re.escape, adict)))

    def one_xlat(match):
        return adict[match.group(0)]

    return rx.sub(one_xlat, text)


def format_unicode(s: str) -> str:
    """Converts a string in CIF text-format to unicode.  Any HTML tags
    contained in the string are removed.  HTML numeric character references
    are unescaped (i.e. converted to unicode).

    Parameters:

    s: string
        The CIF text string to convert

    Returns:

    u: string
        A unicode formatted string.
    """

    s = html.unescape(s)
    s = multiple_replace(s, subs_dict)
    tagclean = re.compile('<.*?>')
    return re.sub(tagclean, '', s)


def handle_subscripts(s: str) -> str:
    s = replace_subscript(s, subscript=True)
    s = replace_subscript(s, subscript=False)
    return s
